package br.ime.usp.mac499.bluetooth.emulator;

import java.rmi.NotBoundException;
import java.rmi.RemoteException;

import br.ime.usp.mac499.bluetooth.emulator.device.EmulatedAdvertisingDevice;
import br.ime.usp.mac499.bluetooth.emulator.device.GenericEmulatedAdvertisingDevice;
import br.ime.usp.mac499.bluetooth.emulator.device.behavior.DiscoveryBehavior;
import br.ime.usp.mac499.bluetooth.emulator.device.behavior.NFCBehavior;
import br.ime.usp.mac499.bluetooth.emulator.device.behavior.ServiceBehavior;
import br.ime.usp.mac499.bluetooth.emulator.device.behavior.TransferBehavior;
import br.ime.usp.mac499.bluetooth.emulator.device.behavior.impl.CorruptTransferBehavior;
import br.ime.usp.mac499.bluetooth.emulator.device.behavior.impl.InvisibleDiscoveryBehavior;
import br.ime.usp.mac499.bluetooth.emulator.device.behavior.impl.NFCTouchBehavior;
import br.ime.usp.mac499.bluetooth.emulator.device.behavior.impl.NormalTransferBehavior;
import br.ime.usp.mac499.bluetooth.emulator.device.behavior.impl.OBEXPushServiceBehavior;
import br.ime.usp.mac499.bluetooth.emulator.device.behavior.impl.RandomDiscoveryBehavior;
import br.ime.usp.mac499.bluetooth.emulator.device.behavior.impl.RejectTransferBehavior;
import br.ime.usp.mac499.bluetooth.emulator.device.behavior.impl.VisibleDiscoveryBehavior;
import br.ime.usp.mac499.bluetooth.simulation.EmulatedDiscoverableMode;
import br.ime.usp.mac499.bluetooth.simulation.IncomingTransferBehavior;

/*
 * This file is part of MAC499 Bluetooth NFC Copyright (C) 2010 Rodolpho Atoji
 * 
 * This program is free software; you can redistribute it and/or modify it under
 * the terms of the GNU General Public License as published by the Free Software
 * Foundation; either version 2 of the License, or (at your option) any later
 * version.
 * 
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
 * details.
 * 
 * You should have received a copy of the GNU General Public License along with
 * this program; if not, write to the Free Software Foundation, Inc., 51
 * Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
 */

/**
 * Factory for building emulated advertising devices.
 * 
 * @author Rodolpho Atoji
 */
public class EmulatedAdvertisingDeviceFactory {

	/**
	 * Assembly a custom emulated advertising device.
	 * 
	 * @param macAddress MAC address (unique).
	 * @param discoverableMode Visible, hidden, random.
	 * @param discoverableAfter Milliseconds after which the device switches
	 * from hidden to visible.
	 * @param randomDiscoveryInterval If random visible, specify interval in
	 * seconds to switch between visible and hidden. The time is chosen
	 * uniformly in this interval. 
	 * @param transferRate Transfer rate in bps.
	 * @param incomingTransfer Accept, reject or corrupt (abort transfer
	 * in a uniform random time between 0 and 30s).
	 * @param sdpEnabled <code>true</code> to have SDP enabled.
	 * @param nfcCapable <code>true</code> to provide NFC touch capability.
	 * @param nfcTouchAfter If NFC-enabled, after how much milliseconds
	 * the device touches the NFC reader.
	 * @return
	 */
	public static EmulatedAdvertisingDevice buildDevice(String macAddress,
			EmulatedDiscoverableMode discoverableMode, long discoverableAfter,
			int randomDiscoveryInterval, int transferRate,
			IncomingTransferBehavior incomingTransfer, boolean sdpEnabled,
			boolean nfcCapable, long nfcTouchAfter) {

		DiscoveryBehavior discoveryBehavior = getDiscoveryBehavior(
				discoverableMode, discoverableAfter, randomDiscoveryInterval);
		TransferBehavior transferBehavior = getTransferBehavior(incomingTransfer);
		ServiceBehavior serviceBehavior = getServiceBehavior(sdpEnabled,
				nfcCapable, nfcTouchAfter);
		NFCBehavior nfcBehavior = getNfcBehavior(macAddress, nfcCapable,
				nfcTouchAfter);

		GenericEmulatedAdvertisingDevice device = new GenericEmulatedAdvertisingDevice(
				macAddress, transferRate, discoveryBehavior, transferBehavior,
				serviceBehavior, nfcBehavior);
		return device;
	}

	private static NFCBehavior getNfcBehavior(String macAddress,
			boolean nfcCapable, long nfcTouchAfter) {
		if (nfcCapable) {
			try {
				return new NFCTouchBehavior(macAddress, nfcTouchAfter);
			} catch (RemoteException e) {
				e.printStackTrace();
			} catch (NotBoundException e) {
				e.printStackTrace();
			}
		}

		return null;
	}

	private static ServiceBehavior getServiceBehavior(boolean sdpEnabled,
			boolean nfcCapable, long nfcTouchAfter) {
		if (sdpEnabled) {
			return new OBEXPushServiceBehavior();
		}

		return null;
	}

	private static TransferBehavior getTransferBehavior(
			IncomingTransferBehavior incomingTransfer) {
		switch (incomingTransfer) {
		case ACCEPT:
			return new NormalTransferBehavior();
		case CORRUPT:
			return new CorruptTransferBehavior();
		case REJECT:
			return new RejectTransferBehavior();
		default:
			return null;
		}
	}

	private static DiscoveryBehavior getDiscoveryBehavior(
			EmulatedDiscoverableMode discoverableMode, long discoverableAfter,
			int randomDiscoveryInterval) {
		switch (discoverableMode) {
		case HIDDEN:
			return new InvisibleDiscoveryBehavior();
		case RANDOM:
			return new RandomDiscoveryBehavior(randomDiscoveryInterval);
		case VISIBLE:
			return new VisibleDiscoveryBehavior(discoverableAfter);
		default:
			return null;
		}
	}

}
