package br.ime.usp.mac499.bluetooth.emulator.device;

import javax.bluetooth.BluetoothStateException;

import br.ime.usp.mac499.bluetooth.emulator.device.behavior.DiscoveryBehavior;
import br.ime.usp.mac499.bluetooth.emulator.device.behavior.NFCBehavior;
import br.ime.usp.mac499.bluetooth.emulator.device.behavior.ServiceBehavior;
import br.ime.usp.mac499.bluetooth.emulator.device.behavior.TransferBehavior;

/*
 * This file is part of MAC499 Bluetooth NFC Copyright (C) 2010 Rodolpho Atoji
 * 
 * This program is free software; you can redistribute it and/or modify it under
 * the terms of the GNU General Public License as published by the Free Software
 * Foundation; either version 2 of the License, or (at your option) any later
 * version.
 * 
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
 * details.
 * 
 * You should have received a copy of the GNU General Public License along with
 * this program; if not, write to the Free Software Foundation, Inc., 51
 * Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
 */

/**
 * Specification of an emulated advertising device.
 * 
 * An emulated device mimics as well as possible the behavior of a
 * Bluetooth-enabled device.
 * 
 * An emulated device can be a combination of several characteristics:
 * 
 * <ul>
 * 	<li><b>Discoverable mode:</b> visible, hidden or random visible;</li>
 * 	<li><b>Discoverable after:</b> a device can remain hidden for some amount
 * 	of time, to simulate arrivals in distinct times;</li>
 * 	<li><b>Transfer rate:</b> data transfer rate is simulated in real time;</li>
 * 	<li><b>Incoming transfer behavior:</b></li> content offering can be
 * 	accepted, rejected or accepted and dropped by transmission error;</li>
 * 	<li><b>SDP presence:</b></li> SDP registry can be available or not on
 * 	a Bluetooth device;</li>
 * <li><b>NFC touch:</b></li> an NFC-enabled device can touch an NFC
 * 	reader after a specified amount of time.</li>
 * </ul>
 * 
 * @author Rodolpho Atoji
 */
public interface EmulatedAdvertisingDevice {

	/**
	 * Initialize the emulated device.
	 * 
	 * @throws BluetoothStateException
	 */
	public abstract void initialize() throws BluetoothStateException;

	/**
	 * Device becomes ready to receive file transfers.
	 */
	public abstract void receivePush();

	/**
	 * Device stops receiving push requests.
	 */
	public abstract void stopPush();

	/**
	 * @return NFC touch behavior.
	 */
	public abstract NFCBehavior getNfcBehavior();

	/**
	 * @return SDP registry behavior.
	 */
	public abstract ServiceBehavior getServiceBehavior();

	/**
	 * @return File transfer behavior.
	 */
	public abstract TransferBehavior getTransferBehavior();

	/**
	 * @return Discovery behavior.
	 */
	public abstract DiscoveryBehavior getDiscoveryBehavior();

	/**
	 * @return MAC address.
	 */
	public abstract String getMacAddress();

	/**
	 * @return Transfer rate in bps.
	 */
	public abstract int getTransferRate();

}