package br.ime.usp.mac499.bluetooth.emulator.device;

import java.io.IOException;

import javax.bluetooth.BluetoothStateException;
import javax.obex.ServerRequestHandler;
import javax.obex.SessionNotifier;

import br.ime.usp.mac499.bluetooth.emulator.device.behavior.DiscoveryBehavior;
import br.ime.usp.mac499.bluetooth.emulator.device.behavior.NFCBehavior;
import br.ime.usp.mac499.bluetooth.emulator.device.behavior.ServiceBehavior;
import br.ime.usp.mac499.bluetooth.emulator.device.behavior.TransferBehavior;

import com.intel.bluetooth.BlueCoveConfigProperties;
import com.intel.bluetooth.BlueCoveImpl;

/*
 * This file is part of MAC499 Bluetooth NFC Copyright (C) 2010 Rodolpho Atoji
 * 
 * This program is free software; you can redistribute it and/or modify it under
 * the terms of the GNU General Public License as published by the Free Software
 * Foundation; either version 2 of the License, or (at your option) any later
 * version.
 * 
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
 * details.
 * 
 * You should have received a copy of the GNU General Public License along with
 * this program; if not, write to the Free Software Foundation, Inc., 51
 * Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
 */

/**
 * Base class for emulated Bluetooth device.
 * 
 * The following behaviors are supported:
 * 
 * <ul>
 * 	<li><b>Discovery:</b> visible, hidden, random;</li>
 * 	<li><b>Transfer:</b> accept, reject, corrupt;</li>
 * 	<li><b>Service:</b> present, absent;</li>
 * 	<li><b>NFC:</b> present, absent.</li>
 * </ul>
 * 
 * Each behavior can have its own parameters.
 * 
 * @author Rodolpho Atoji
 */
public class GenericEmulatedAdvertisingDevice implements
		EmulatedAdvertisingDevice {

	protected static final String BTGOEP_CONNECTION_URL = "btgoep://localhost:5;authenticate=false;master=false;encrypt=false";

	protected final int transferRate;

	protected final String macAddress;

	protected final DiscoveryBehavior discoveryBehavior;

	protected final TransferBehavior transferBehavior;

	protected final ServiceBehavior serviceBehavior;

	protected final NFCBehavior nfcBehavior;
	
	public GenericEmulatedAdvertisingDevice(String macAddress,
			int transferRate, DiscoveryBehavior discoveryBehavior,
			TransferBehavior transferBehavior, ServiceBehavior serviceBehavior,
			NFCBehavior nfcBehavior) {
		super();
		this.macAddress = macAddress;
		this.transferRate = transferRate;
		this.discoveryBehavior = discoveryBehavior;
		this.transferBehavior = transferBehavior;
		this.serviceBehavior = serviceBehavior;
		this.nfcBehavior = nfcBehavior;
	}

	@Override
	public void initialize() throws BluetoothStateException {
		BlueCoveImpl.setConfigProperty(BlueCoveConfigProperties.PROPERTY_STACK,
				BlueCoveImpl.STACK_EMULATOR);
		BlueCoveImpl.setConfigProperty(
				BlueCoveConfigProperties.PROPERTY_EMULATOR_PORT, "0");
		BlueCoveImpl.setConfigProperty(
				BlueCoveConfigProperties.PROPERTY_LOCAL_DEVICE_ADDRESS,
				macAddress);
		BlueCoveImpl.useThreadLocalBluetoothStack();

		Object currentAdapter = BlueCoveImpl.getThreadBluetoothStackID();
		BlueCoveImpl.setThreadBluetoothStackID(currentAdapter);

		discoveryBehavior.initialize();

		if (serviceBehavior != null)
				serviceBehavior.initialize(BTGOEP_CONNECTION_URL);

		transferBehavior.initialize(transferRate);
		
		if (nfcBehavior != null)
			nfcBehavior.initialize();
	}

	@Override
	public void receivePush() {
		if (serviceBehavior == null)
			return;
		if (transferBehavior == null)
			return;

		SessionNotifier sessionNotifier = serviceBehavior.getNotifier();
		ServerRequestHandler requestHandler = transferBehavior
				.getRequestHandler();

		try {
			sessionNotifier.acceptAndOpen(requestHandler);
		} catch (IOException e) {
		}
	}

	@Override
	public void stopPush() {
		if (serviceBehavior == null)
			return;
		
		SessionNotifier sessionNotifier = serviceBehavior.getNotifier();
		try {
			sessionNotifier.close();
		} catch (IOException e) {
			e.printStackTrace();
		}
	}

	@Override
	public int getTransferRate() {
		return transferRate;
	}

	@Override
	public String getMacAddress() {
		return macAddress;
	}

	@Override
	public DiscoveryBehavior getDiscoveryBehavior() {
		return discoveryBehavior;
	}

	@Override
	public TransferBehavior getTransferBehavior() {
		return transferBehavior;
	}

	@Override
	public ServiceBehavior getServiceBehavior() {
		return serviceBehavior;
	}

	@Override
	public NFCBehavior getNfcBehavior() {
		return nfcBehavior;
	}
	
}
