package br.ime.usp.mac499.bluetooth.simulation;

import java.io.File;
import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.IOException;
import java.util.Properties;

import br.ime.usp.mac499.bluetooth.emulator.EmulatedAdvertisingDeviceFactory;
import br.ime.usp.mac499.bluetooth.emulator.device.EmulatedAdvertisingDevice;

/*
 * This file is part of MAC499 Bluetooth NFC Copyright (C) 2010 Rodolpho Atoji
 * 
 * This program is free software; you can redistribute it and/or modify it under
 * the terms of the GNU General Public License as published by the Free Software
 * Foundation; either version 2 of the License, or (at your option) any later
 * version.
 * 
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
 * details.
 * 
 * You should have received a copy of the GNU General Public License along with
 * this program; if not, write to the Free Software Foundation, Inc., 51
 * Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
 */

/**
 * Loads and parses a properties-formatted simulation scenario.
 * 
 * @author Rodolpho Atoji
 */
public class PropertiesSimulationScenarioParser implements
		SimulationScenarioParser {

	private static final int MAX_ALLOWED_DEVICES = 512;

	private static final long DEFAULT_DISCOVERABLE_AFTER = 0;

	private static final int DEFAULT_BPS = 2500000;

	private static final boolean DEFAULT_NFC_CAPABLE = false;

	private static final long DEFAULT_NFC_TOUCH_AFTER = 0;

	private static final boolean DEFAULT_SDP_ENABLED = true;

	private static final int DEFAULT_RANDOM_INTERVAL = 0;

	public SimulationScenario loadScenario(Properties properties) {
		if (properties == null)
			return null;

		return parseScenario(properties);
	}

	@Override
	public SimulationScenario loadScenario(String filePath) {
		File scenarioFile = new File(filePath);

		if (!scenarioFile.exists())
			return null;

		try {
			FileInputStream inputStream = new FileInputStream(scenarioFile);
			Properties propertiesScenario = new Properties();
			propertiesScenario.load(inputStream);

			return parseScenario(propertiesScenario);
		} catch (FileNotFoundException e) {
			return null;
		} catch (IOException e) {
			e.printStackTrace();
		}

		return null;
	}

	private SimulationScenario parseScenario(Properties propertiesScenario) {
		SimulationScenario simulationScenario = new SimulationScenario();

		parseSimulationParameters(simulationScenario, propertiesScenario);
		parseAdvertisingDevices(simulationScenario, propertiesScenario);

		return simulationScenario;
	}

	private void parseAdvertisingDevices(SimulationScenario simulationScenario,
			Properties propertiesScenario) {
		String currentDevice;
		String macAddress;
		EmulatedAdvertisingDevice device;

		for (int i = 0; i < MAX_ALLOWED_DEVICES; i++) {
			currentDevice = "device." + i;
			macAddress = propertiesScenario.getProperty(currentDevice);

			if (macAddress == null)
				break;

			device = parseAdvertisingDevice(currentDevice, propertiesScenario);
			simulationScenario.addAdvertisingDevice(device);
		}
	}

	private EmulatedAdvertisingDevice parseAdvertisingDevice(
			String currentDevice, Properties scenario) {
		String macAddress = scenario.getProperty(currentDevice);
		EmulatedDiscoverableMode discoverableMode = parseDiscoverableMode(
				currentDevice, scenario);
		long discoverableAfter = parseDiscoverableAfter(currentDevice, scenario);
		int transferRate = parseTransferRate(currentDevice, scenario);
		int randomDiscoveryInterval = parseRandomDiscoveryInterval(
				currentDevice, scenario);
		boolean nfcCapable = parseNfcCapable(currentDevice, scenario);
		long nfcTouchAfter = parseNfcTouchAfter(currentDevice, scenario);
		IncomingTransferBehavior incomingTransfer = parseIncomingTransfer(
				currentDevice, scenario);
		boolean sdpEnabled = parseSdpEnabled(currentDevice, scenario);

		EmulatedAdvertisingDevice device = EmulatedAdvertisingDeviceFactory
				.buildDevice(macAddress, discoverableMode, discoverableAfter,
						randomDiscoveryInterval, transferRate,
						incomingTransfer, sdpEnabled, nfcCapable, nfcTouchAfter);

		return device;
	}

	private int parseRandomDiscoveryInterval(String currentDevice,
			Properties scenario) {
		String randomInterval = scenario.getProperty(currentDevice
				+ ".random_discovery");
		int randomDiscoveryInterval;
		try {
			randomDiscoveryInterval = Integer.parseInt(randomInterval);
		} catch (Exception e) {
			randomDiscoveryInterval = DEFAULT_RANDOM_INTERVAL;
		}

		return randomDiscoveryInterval;
	}

	private boolean parseSdpEnabled(String currentDevice, Properties scenario) {
		String sdpEnabled = scenario
				.getProperty(currentDevice + ".sdp_enabled");
		boolean sdpEnabledBoolean;
		try {
			sdpEnabledBoolean = Boolean.parseBoolean(sdpEnabled);
		} catch (Exception e) {
			sdpEnabledBoolean = DEFAULT_SDP_ENABLED;
		}

		return sdpEnabledBoolean;
	}

	private IncomingTransferBehavior parseIncomingTransfer(
			String currentDevice, Properties scenario) {
		String incomingTransfer = scenario.getProperty(currentDevice
				+ ".incoming_transfer");
		IncomingTransferBehavior transferBehavior = IncomingTransferBehavior
				.getMapping(incomingTransfer);

		return transferBehavior;
	}

	private long parseNfcTouchAfter(String currentDevice, Properties scenario) {
		String nfcTouchAfter = scenario.getProperty(currentDevice
				+ ".nfc_touch_after");
		long nfcTouchAfterMillis;
		try {
			nfcTouchAfterMillis = Integer.parseInt(nfcTouchAfter) * 1000;
		} catch (Exception e) {
			nfcTouchAfterMillis = DEFAULT_NFC_TOUCH_AFTER;
		}

		return nfcTouchAfterMillis;
	}

	private boolean parseNfcCapable(String currentDevice, Properties scenario) {
		String nfcCapable = scenario
				.getProperty(currentDevice + ".nfc_capable");
		boolean nfcCapableBoolean;
		try {
			nfcCapableBoolean = Boolean.parseBoolean(nfcCapable);
		} catch (Exception e) {
			nfcCapableBoolean = DEFAULT_NFC_CAPABLE;
		}

		return nfcCapableBoolean;
	}

	private int parseTransferRate(String currentDevice, Properties scenario) {
		String transferRate = scenario.getProperty(currentDevice
				+ ".transfer_rate");
		int transferRateBps;
		try {
			transferRateBps = Integer.parseInt(transferRate);
		} catch (Exception e) {
			transferRateBps = DEFAULT_BPS;
		}

		return transferRateBps;
	}

	private long parseDiscoverableAfter(String currentDevice,
			Properties scenario) {
		String discoverableAfter = scenario.getProperty(currentDevice
				+ ".discoverable_after");
		long discoverableAfterMillis;
		try {
			discoverableAfterMillis = Integer.parseInt(discoverableAfter) * 1000;
		} catch (Exception e) {
			discoverableAfterMillis = DEFAULT_DISCOVERABLE_AFTER;
		}

		return discoverableAfterMillis;
	}

	private EmulatedDiscoverableMode parseDiscoverableMode(
			String currentDevice, Properties scenario) {
		String discoverableMode = scenario.getProperty(currentDevice
				+ ".discoverable_mode");
		EmulatedDiscoverableMode emulatedDiscoverableMode = EmulatedDiscoverableMode
				.getMapping(discoverableMode);

		return emulatedDiscoverableMode;
	}

	private void parseSimulationParameters(
			SimulationScenario simulationScenario, Properties propertiesScenario) {
		String simulationName;
		int simulationDuration;

		simulationName = propertiesScenario.getProperty("simulation.name",
				"Default name");

		String durationString = propertiesScenario
				.getProperty("simulation.duration");
		try {
			simulationDuration = Integer.parseInt(durationString);
		} catch (Exception e) {
			simulationDuration = 0;
		}

		simulationScenario.setDuration(simulationDuration);
		simulationScenario.setName(simulationName);
	}

}
