package br.ime.usp.mac499.nfc.impl;

import java.io.FileReader;
import java.io.IOException;
import java.rmi.RemoteException;
import java.util.List;
import java.util.Properties;
import java.util.concurrent.Semaphore;

import javax.microedition.contactless.ndef.NDEFMessage;
import javax.microedition.contactless.ndef.NDEFRecord;
import javax.microedition.contactless.ndef.NDEFRecordListener;
import javax.microedition.contactless.ndef.NDEFRecordType;

/*
 * This file is part of MAC499 Bluetooth NFC 
 * Copyright (C) 2010 Rodolpho Atoji
 * 
 * This program is free software; you can redistribute it and/or modify it under
 * the terms of the GNU General Public License as published by the Free Software
 * Foundation; either version 2 of the License, or (at your option) any later
 * version.
 * 
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
 * details.
 * 
 * You should have received a copy of the GNU General Public License along with
 * this program; if not, write to the Free Software Foundation, Inc., 51
 * Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
 */

/**
 * RMI emulation of NFC touches.
 * 
 * When an emulated device calls <code>notifyTouch()</code> via RMI,
 * all <code>NDEFRecordListener</code>s are notified of a new NFC touch.
 * 
 * The number of emulated NFC readers can be set from the file
 * <code>nfctouch.properties</code>.
 * 
 * @author Rodolpho Atoji
 */
public class NDEFDiscoveryManagerRMI implements NFCTouchNotify {

	private final static String CONFIG_PATH = "./config/nfctouch.properties";

	private final List<NDEFRecordListener> ndefListenerList;

	private final Semaphore nfcSemaphore;

	public NDEFDiscoveryManagerRMI(List<NDEFRecordListener> ndefListenerList) {
		this.ndefListenerList = ndefListenerList;

		Properties configProperties = loadConfiguration();
		int readers = 1;

		try {
			String readersString = configProperties.getProperty("nfc.readers");
			readers = Integer.parseInt(readersString);
			System.out.println("[NFC EMULATOR] " + readers
					+ " reader(s) available");
		} catch (Exception e) {
			readers = 1;
			System.err.println("[NFC EMULATOR] Config file not found, using 1"
					+ " reader as default");
		}

		this.nfcSemaphore = new Semaphore(readers);
	}

	private Properties loadConfiguration() {
		Properties configProperties = new Properties();

		try {
			FileReader fileReader = new FileReader(CONFIG_PATH);
			configProperties.load(fileReader);
		} catch (IOException e) {
			e.printStackTrace();
			System.err.println("[ERROR] Cannot load configuration for NFC");
			return null;
		}

		return configProperties;
	}

	@Override
	public void notifyTouch(String macAddress, String connectionURL)
			throws RemoteException {
		try {
			nfcSemaphore.acquire();

			NDEFRecordType ndefRecordType = new NDEFRecordType(
					NDEFRecordType.MIME, "text/plain");
			NDEFRecord ndefRecord = new NDEFRecord(ndefRecordType, null,
					connectionURL.getBytes());
			NDEFMessage ndefMessage = new NDEFMessage(
					new NDEFRecord[] { ndefRecord });

			/* At this point the main software receives a touch notification */
			for (NDEFRecordListener listener : ndefListenerList)
				listener.recordDetected(ndefMessage);
		} catch (InterruptedException e) {
			e.printStackTrace();
		} finally {
			nfcSemaphore.release();
		}
	}

}
