package javax.microedition.contactless;

import java.rmi.NotBoundException;
import java.rmi.RemoteException;
import java.rmi.registry.LocateRegistry;
import java.rmi.registry.Registry;
import java.rmi.server.UnicastRemoteObject;
import java.util.HashMap;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;

import javax.microedition.contactless.ndef.NDEFRecordListener;
import javax.microedition.contactless.ndef.NDEFRecordType;

import br.ime.usp.mac499.nfc.impl.NDEFDiscoveryManagerRMI;
import br.ime.usp.mac499.nfc.impl.NFCTouchNotify;

/*
 * This file is part of MAC499 Bluetooth NFC Copyright (C) 2010 Rodolpho Atoji
 * 
 * This program is free software; you can redistribute it and/or modify it under
 * the terms of the GNU General Public License as published by the Free Software
 * Foundation; either version 2 of the License, or (at your option) any later
 * version.
 * 
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
 * details.
 * 
 * You should have received a copy of the GNU General Public License along with
 * this program; if not, write to the Free Software Foundation, Inc., 51
 * Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
 */

/**
 * Implementation of JSR 257 NFC <code>DiscoveryManager</code>.
 * 
 * Only NDEF records are supported.
 * 
 * @author Rodolpho Atoji
 */
public class DiscoveryManager {

	private static final int RMI_PORT = 1099;

	private static final String RMI_NAME_BIND = "NFCTouch";

	private final Map<Integer, List<NDEFRecordListener>> ndefListenerMap;

	private static DiscoveryManager discoveryManager;

	private NDEFDiscoveryManagerRMI ndefDiscoveryManagerRMI;

	private Registry rmiRegistry;

	public static DiscoveryManager getInstance() {
		synchronized (DiscoveryManager.class) {
			if (discoveryManager == null)
				discoveryManager = new DiscoveryManager();
			return discoveryManager;
		}
	}

	private DiscoveryManager() {
		this.ndefListenerMap = new HashMap<Integer, List<NDEFRecordListener>>();

		ndefListenerMap.put(NDEFRecordType.EMPTY,
				new LinkedList<NDEFRecordListener>());
		ndefListenerMap.put(NDEFRecordType.EXTERNAL_RTD,
				new LinkedList<NDEFRecordListener>());
		ndefListenerMap.put(NDEFRecordType.MIME,
				new LinkedList<NDEFRecordListener>());
		ndefListenerMap.put(NDEFRecordType.NFC_FORUM_RTD,
				new LinkedList<NDEFRecordListener>());
		ndefListenerMap.put(NDEFRecordType.UNKNOWN,
				new LinkedList<NDEFRecordListener>());
		ndefListenerMap.put(NDEFRecordType.URI,
				new LinkedList<NDEFRecordListener>());
	}

	private synchronized void initializeRMI() {
		List<NDEFRecordListener> ndefListenerList = ndefListenerMap
				.get(NDEFRecordType.MIME);
		this.ndefDiscoveryManagerRMI = new NDEFDiscoveryManagerRMI(
				ndefListenerList);

		try {
			System.setProperty("java.rmi.server.hostname", "localhost");
			NFCTouchNotify touchNotify = (NFCTouchNotify) UnicastRemoteObject
					.exportObject(ndefDiscoveryManagerRMI, RMI_PORT);

			this.rmiRegistry = LocateRegistry.createRegistry(RMI_PORT);
			rmiRegistry.bind(RMI_NAME_BIND, touchNotify);
		} catch (Exception e) {
			e.printStackTrace();
		}
	}

	private synchronized void shutdownRMI() {
		try {
			UnicastRemoteObject.unexportObject(ndefDiscoveryManagerRMI, true);
			Registry registry = LocateRegistry.getRegistry(RMI_PORT);
			registry.unbind(RMI_NAME_BIND);
			UnicastRemoteObject.unexportObject(rmiRegistry, true);
		} catch (RemoteException e) {
			e.printStackTrace();
		} catch (NotBoundException e) {
			e.printStackTrace();
		}
	}

	public void addTargetListener(TargetListener listener, TargetType targetType)
			throws ContactlessException, IllegalStateException {

	}

	public void addNDEFRecordListener(NDEFRecordListener listener,
			NDEFRecordType recordType) throws ContactlessException {
		if (listener == null)
			throw new NullPointerException("Listener is null");
		if (recordType == null)
			throw new NullPointerException("RecordType is null");

		List<NDEFRecordListener> listenerList = ndefListenerMap.get(recordType
				.getFormat());

		if (listenerList != null) {
			if (listenerList.isEmpty())
				initializeRMI();

			listenerList.add(listener);
		} else {
			throw new ContactlessException("Record type "
					+ recordType.getFormat() + " not supported");
		}
	}

	public void removeTargetListener(TargetListener listener,
			TargetType targetType) {
	}

	public void removeNDEFRecordListener(NDEFRecordListener listener,
			NDEFRecordType recordType) {
		List<NDEFRecordListener> listenerList = ndefListenerMap.get(recordType
				.getFormat());

		if (listenerList != null) {
			listenerList.remove(listener);

			if (listenerList.isEmpty())
				shutdownRMI();
		}
	}

	public void addTransactionListener(TransactionListener listener)
			throws ContactlessException {

	}

	public void removeTransactionListener(TransactionListener listener) {

	}

	public static TargetType[] getSupportedTargetTypes() {
		return null;
	}

	public void setProperty(String name, String value) {

	}

	public String getProperty(java.lang.String name) {
		return null;
	}

}
