package br.ime.usp.mac499.bluetooth;

import javax.bluetooth.BluetoothStateException;

import br.ime.usp.mac499.bluetooth.advertising.AdvertisingManager;
import br.ime.usp.mac499.bluetooth.advertising.RunnableAdvertiser;
import br.ime.usp.mac499.bluetooth.common.BluetoothAdapterMonitor;
import br.ime.usp.mac499.bluetooth.common.BluetoothAdapterScheduler;
import br.ime.usp.mac499.bluetooth.common.BluetoothResources;
import br.ime.usp.mac499.bluetooth.common.GenericBluetoothAdapterScheduler;
import br.ime.usp.mac499.bluetooth.common.config.BluetoothParameters;
import br.ime.usp.mac499.bluetooth.common.config.PropertiesBluetoothParameters;
import br.ime.usp.mac499.bluetooth.common.discovery.DeviceDiscoverer;
import br.ime.usp.mac499.bluetooth.common.discovery.DiscoveryObserver;
import br.ime.usp.mac499.bluetooth.common.discovery.DiscoveryResultProcessor;
import br.ime.usp.mac499.bluetooth.common.discovery.DiscoveryType;
import br.ime.usp.mac499.bluetooth.common.discovery.GenericDeviceDiscoverer;
import br.ime.usp.mac499.bluetooth.common.sdp.ServiceSearchDispatcher;
import br.ime.usp.mac499.bluetooth.common.stack.BluecoveLocalAdaptersManager;
import br.ime.usp.mac499.bluetooth.common.stack.LocalAdaptersManager;
import br.ime.usp.mac499.bluetooth.db.DatabaseException;
import br.ime.usp.mac499.bluetooth.db.DatabaseManager;
import br.ime.usp.mac499.bluetooth.db.DiscoveryMetadataObserver;
import br.ime.usp.mac499.bluetooth.db.HSQLDatabaseManager;
import br.ime.usp.mac499.bluetooth.model.AdvertisingDeviceManager;
import br.ime.usp.mac499.bluetooth.model.MediaManager;
import br.ime.usp.mac499.bluetooth.model.MultimediaMultipleFileManager;
import br.ime.usp.mac499.nfc.NFCDeviceManager;

/*
 * This file is part of MAC499 Bluetooth NFC
 * Copyright (C) 2010 Rodolpho Atoji
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  
 * 02110-1301, USA.
 */

/**
 * Launching class for Bluetooth marketing software.
 * 
 * Bluetooth adapters required.
 * 
 * @author Rodolpho Atoji
 */
public class BluetoothMarketing {

	private final static String FLAG_MEDIAFILE = "-mediaFile";

	public static void main(String[] args) {
		initializeDatabaseManager();

		LocalAdaptersManager adapterManager = initializeLocalAdapters();
		BluetoothParameters parameters = new PropertiesBluetoothParameters();
		BluetoothAdapterMonitor monitor = new BluetoothAdapterMonitor(
				adapterManager.getTotalAdapters(),
				parameters.getDiscoveryAdapterNumber(),
				parameters.getMaxConcurrentTransfers());
		AdvertisingDeviceManager deviceManager = new AdvertisingDeviceManager();

		MediaManager mediaManager = parseCommandLine(args);
		BluetoothResources.initialize(adapterManager, monitor, parameters,
				deviceManager, mediaManager);

		BluetoothAdapterScheduler scheduler = initializeAdapterScheduler();
		initializeDiscoveryProcess(deviceManager, scheduler);
		initializeAdvertisingProcess(scheduler);
		initializeNFCDeviceManager();
	}

	private static BluetoothAdapterScheduler initializeAdapterScheduler() {
		BluetoothAdapterScheduler scheduler = null;
		try {
			scheduler = new GenericBluetoothAdapterScheduler();
		} catch (Exception e) {
			System.err
					.println("[ERROR] Cannot initialize Bluetooth adapters, aborting");
			System.exit(1);
		}
		return scheduler;
	}

	private static MediaManager parseCommandLine(String[] args) {
		if (args == null || args.length == 0) {
			System.err.println("[WARN] Trying to load default media...");
			return initializeMedia();
		}

		MediaManager mediaManager = new MultimediaMultipleFileManager();

		for (int i = 0; i < args.length; i++) {
			if (args[i].startsWith(FLAG_MEDIAFILE))
				parseMediaFile(mediaManager, args[i]);
			else
				System.err.println("[WARN] Ignored flag: " + args[i]);
		}

		return mediaManager;
	}

	private static void parseMediaFile(MediaManager mediaManager, String arg) {
		int separatorIndex = arg.indexOf("=");
		String mediaList = arg.substring(separatorIndex + 1);
		String pathList[] = mediaList.split(",");

		for (String path : pathList)
			mediaManager.reloadMedia(path);
	}

	private static MediaManager initializeMedia() {
		MediaManager mediaManager = new MultimediaMultipleFileManager();
		mediaManager.reloadMedia("media/logo-palmeiras.jpg");
		mediaManager.reloadMedia("media/hino-original-low.mp3");
		return mediaManager;
	}

	private static LocalAdaptersManager initializeLocalAdapters() {
		LocalAdaptersManager adapterManager = new BluecoveLocalAdaptersManager();
		try {
			adapterManager.initialize();
		} catch (BluetoothStateException e) {
			System.err
					.println("[ERROR] Cannot initialize Bluetooth adapters, aborting");
			System.exit(1);
		}
		return adapterManager;
	}

	private static void initializeDatabaseManager() {
		DatabaseManager databaseManager = new HSQLDatabaseManager();
		try {
			databaseManager.open();
		} catch (DatabaseException e) {
			System.err.println("[ERROR] Cannot open database, aborting");
			System.exit(1);
		}
	}

	private static void initializeNFCDeviceManager() {
		NFCDeviceManager nfcDeviceManager = new NFCDeviceManager();
		nfcDeviceManager.start();
	}

	private static void initializeAdvertisingProcess(
			BluetoothAdapterScheduler scheduler) {
		RunnableAdvertiser advertiser = new RunnableAdvertiser(scheduler);
		AdvertisingManager advertisingManager = new AdvertisingManager(
				advertiser);
		advertisingManager.startAdvertising();
	}

	private static void initializeDiscoveryProcess(
			AdvertisingDeviceManager deviceManager,
			BluetoothAdapterScheduler scheduler) {
		DeviceDiscoverer deviceDiscoverer = GenericDeviceDiscoverer
				.getInstance();
		deviceDiscoverer = GenericDeviceDiscoverer.getInstance();

		DiscoveryObserver discoveryObserver = new DiscoveryResultProcessor(
				deviceManager);
		deviceDiscoverer.addDiscoveryObserver(discoveryObserver);

		DiscoveryObserver dispatcher = new ServiceSearchDispatcher(
				deviceManager, scheduler);
		deviceDiscoverer.addDiscoveryObserver(dispatcher);

		DiscoveryMetadataObserver metadataObserver = new DiscoveryMetadataObserver();
		deviceDiscoverer.addDiscoveryObserver(metadataObserver);

		scheduler
				.registerDiscoveryTask(deviceDiscoverer, DiscoveryType.GENERAL);
	}

}
