package br.ime.usp.mac499.bluetooth.advertising;

import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.IOException;
import java.util.HashMap;
import java.util.Properties;

import br.ime.usp.mac499.bluetooth.model.AdvertisingDevice;

/*
 * This file is part of MAC499 Bluetooth NFC 
 * Copyright (C) 2010 Rodolpho Atoji
 * 
 * This program is free software; you can redistribute it and/or modify it under
 * the terms of the GNU General Public License as published by the Free Software
 * Foundation; either version 2 of the License, or (at your option) any later
 * version.
 * 
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
 * details.
 * 
 * You should have received a copy of the GNU General Public License along with
 * this program; if not, write to the Free Software Foundation, Inc., 51
 * Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
 */

/**
 * Filter each target Bluetooth device based on its MAC address.
 * 
 * @author Rodolpho Atoji
 */
public class AdvertisingDeviceAddressFilter implements
		AdvertisingDeviceFilterChain {

	private AdvertisingDeviceFilterChain nextFilter;

	/**
	 * Map of filtered devices (allowed or blacklisted).
	 */
	private HashMap<String, FilterAction> filteredDevicesMap;

	/**
	 * Current filter running policy (allow all, filter, deny all).
	 */
	private FilterMode filterMode;

	/**
	 * Available actions to take for each MAC address.
	 */
	private final static FilterAction[] filterActions = { FilterAction.ALLOWED,
			FilterAction.BLACKLISTED };

	/**
	 * Provided a filter file, constructs a filter based on its rules.
	 * 
	 * A filter file can filter out all devices, allow specific devices by MAC 
	 * address, or allow all devices. 
	 * 
	 * @param filterFilePath Path to filter file.
	 */
	public AdvertisingDeviceAddressFilter(String filterFilePath) {
		this.filteredDevicesMap = new HashMap<String, FilterAction>();

		buildFilterMap(filterFilePath);
	}

	private void buildFilterMap(String filterFilePath) {
		Properties filterProperties = readFilterFileFromDisk(filterFilePath);

		FilterMode filterMode = determineFilterOperationMode(filterProperties);
		this.filterMode = filterMode;

		if (filterMode != FilterMode.DO_FILTER)
			return;

		for (FilterAction filterAction : filterActions)
			loadFilteredDevicesIntoMap(filterProperties,
					filterAction.getRelatedProperty(), filterAction);
	}

	private FilterMode determineFilterOperationMode(Properties filterProperties) {
		String filterMode = filterProperties.getProperty("filter.mode");
		return FilterMode.getMappingForProperty(filterMode);
	}

	private void loadFilteredDevicesIntoMap(Properties filterProperties,
			String targetProperty, FilterAction targetAction) {
		boolean remainingDevicesToFilter = true;
		String currentMacToFilter;
		for (int i = 0; remainingDevicesToFilter; i++) {
			currentMacToFilter = filterProperties.getProperty(targetProperty
					+ i);

			if (currentMacToFilter == null)
				remainingDevicesToFilter = false;
			else
				filteredDevicesMap.put(currentMacToFilter, targetAction);
		}
	}

	private Properties readFilterFileFromDisk(String filterFilePath) {
		Properties filterProperties = new Properties();

		FileInputStream fileInputStream;
		try {
			fileInputStream = new FileInputStream(filterFilePath);
			filterProperties.load(fileInputStream);
		} catch (FileNotFoundException e) {
			System.err.println("[WARN] Device filter file not found: "
					+ filterFilePath + ", using defaults");
		} catch (IOException e) {
			System.err.println("[ERROR] Failed to read filter file: "
					+ filterFilePath);
		}

		return filterProperties;
	}

	@Override
	public void setNext(AdvertisingDeviceFilterChain processor) {
		this.nextFilter = processor;
	}

	@Override
	public boolean shouldFilter(AdvertisingDevice device) {
		if (filterMode == FilterMode.ALLOW_ALL) {
			if (nextFilter != null)
				return nextFilter.shouldFilter(device);
			return false;
		} else if (filterMode == FilterMode.DENY_ALL)
			return true;

		String macAddress = device.getMacAddress();
		FilterAction filterAction = filteredDevicesMap.get(macAddress);

		if (filterAction == null) {
			return true;
		} else if (filterAction == FilterAction.BLACKLISTED)
			return true;
		else if (nextFilter != null)
			return nextFilter.shouldFilter(device);

		return false;
	}

}
