package br.ime.usp.mac499.bluetooth.advertising;

import br.ime.usp.mac499.bluetooth.common.BluetoothResources;
import br.ime.usp.mac499.bluetooth.common.config.BluetoothParameters;
import br.ime.usp.mac499.bluetooth.common.sdp.ServiceStatus;
import br.ime.usp.mac499.bluetooth.model.AdvertisingDevice;
import br.ime.usp.mac499.bluetooth.model.MediaManager;

/*
 * This file is part of MAC499 Bluetooth NFC 
 * Copyright (C) 2010 Rodolpho Atoji
 * 
 * This program is free software; you can redistribute it and/or modify it under
 * the terms of the GNU General Public License as published by the Free Software
 * Foundation; either version 2 of the License, or (at your option) any later
 * version.
 * 
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
 * details.
 * 
 * You should have received a copy of the GNU General Public License along with
 * this program; if not, write to the Free Software Foundation, Inc., 51
 * Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
 */

/**
 * Link of filter chain which takes care of advertising limits to a
 * target device.
 * 
 * This filter takes out a device of an advertising process if it:
 * 
 * <ul>
 * 	<li>Is currently in a transfer;</li>
 * 	<li>Has exceeded maximum failed advertising retries;</li>
 * 	<li>Has exceeded maximum successful received contents;</li>
 * 	<li>Has no advertising content loaded.</li>
 * </ul>
 * 
 * @author Rodolpho Atoji
 */
public class AdvertisingMediaFilter implements AdvertisingDeviceFilterChain {

	private AdvertisingDeviceFilterChain nextFilter;

	private final BluetoothParameters parameters;
	
	private final MediaManager mediaManager;

	/**
	 * Construct this link as the end of a chain.
	 * 
	 * @param parameters System-wide parameters.
	 */
	public AdvertisingMediaFilter(BluetoothParameters parameters) {
		this.parameters = parameters;
		this.mediaManager = BluetoothResources.getAdvertisingMediaManager();
	}

	/**
	 * Constructs this link as part of a chain.
	 * 
	 * @param parameters System-wide parameters.
	 * @param processor Next link in the chain.
	 */
	public AdvertisingMediaFilter(BluetoothParameters parameters,
			AdvertisingDeviceFilterChain processor) {
		this(parameters);
		this.nextFilter = processor;
	}

	@Override
	public void setNext(AdvertisingDeviceFilterChain processor) {
		this.nextFilter = processor;
	}

	@Override
	public boolean shouldFilter(AdvertisingDevice device) {
		if (device.getServiceSearchStatus() != ServiceStatus.COMPLETED)
			return true;
				
		int maxRetries = parameters.getMaxAdvertisingRetries();
		int maxSuccesses = parameters.getMaxAdvertisingSuccesses();
		
		if (device.isLockedForCommunication())
			return true;
		if (device.getFailedAdvertisementAttempts() >= maxRetries)
			return true;
		if (device.getSuccesfulAdvertisementAttempts() >= maxSuccesses)
			return true;
		if (mediaManager.getNextMediaForDevice(device) == null)
			return true;
		
		if (nextFilter != null)
			return nextFilter.shouldFilter(device);

		return false;
	}

}
