package br.ime.usp.mac499.bluetooth.advertising;

import java.util.LinkedList;
import java.util.List;

import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import br.ime.usp.mac499.bluetooth.common.BluetoothAdapterScheduler;
import br.ime.usp.mac499.bluetooth.common.BluetoothResources;
import br.ime.usp.mac499.bluetooth.db.AdvertisingMetadataObserver;
import br.ime.usp.mac499.bluetooth.io.FilePusher;
import br.ime.usp.mac499.bluetooth.io.OBEXFilePusher;
import br.ime.usp.mac499.bluetooth.model.AdvertisingDevice;
import br.ime.usp.mac499.bluetooth.model.AdvertisingDeviceManager;
import br.ime.usp.mac499.bluetooth.model.AdvertisingMediaFile;
import br.ime.usp.mac499.bluetooth.model.DefaultPriorityPolicy;
import br.ime.usp.mac499.bluetooth.model.MediaManager;
import br.ime.usp.mac499.bluetooth.model.Priority;
import br.ime.usp.mac499.bluetooth.model.PriorityPolicy;
import br.ime.usp.mac499.util.PathUtils;
import br.ime.usp.mac499.util.StringUtils;

/*
 * This file is part of MAC499 Bluetooth NFC 
 * Copyright (C) 2010 Rodolpho Atoji
 * 
 * This program is free software; you can redistribute it and/or modify it under
 * the terms of the GNU General Public License as published by the Free Software
 * Foundation; either version 2 of the License, or (at your option) any later
 * version.
 * 
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
 * details.
 * 
 * You should have received a copy of the GNU General Public License along with
 * this program; if not, write to the Free Software Foundation, Inc., 51
 * Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
 */

/**
 * Runnable task of an advertising process.
 * 
 * An advertising process takes care of:
 * 
 * <ul>
 * 	<li>Consuming devices discovered by discovery process;</li>
 * 	<li>Dispatching OBEX file pusher tasks for file transfers;</li>
 * 	<li>Prioritizing devices advertisement order;</li>
 * 	<li>Filtering out devices which should not be advertised.</li>
 * <ul>	 
 * 
 * @author Rodolpho Atoji
 */
public class RunnableAdvertiser implements Runnable {

	private final AdvertisingDeviceManager deviceManager;

	private final AdvertisingDeviceFilterChain filterChain;

	private final MediaManager mediaManager;

	private final BluetoothAdapterScheduler scheduler;

	private final Logger logger;

	/**
	 * Builds an advertiser with the proper adapter scheduler, which is 
	 * responsible of assigning Bluetooth adapters for file transfers. 
	 * 
	 * @param scheduler Adapter scheduler.
	 */
	public RunnableAdvertiser(BluetoothAdapterScheduler scheduler) {
		this.deviceManager = BluetoothResources.getAdvertisingDeviceManager();
		this.mediaManager = BluetoothResources.getAdvertisingMediaManager();
		this.scheduler = scheduler;
		this.logger = LoggerFactory.getLogger(RunnableAdvertiser.class);

		String filterFilePath = PathUtils.getDeviceFilterPath();
		this.filterChain = new AdvertisingDeviceAddressFilter(filterFilePath);
		AdvertisingMediaFilter mediaFilter = new AdvertisingMediaFilter(
				BluetoothResources.getParameters());
		filterChain.setNext(mediaFilter);
	}

	@Override
	public void run() {
		/*
		 * The order in which the priorities are specified here define the
		 * order the target devices will be preferred in the advertising
		 * process.
		 */
		PriorityPolicy policy = new DefaultPriorityPolicy(
				Priority.NFC_INITIATED, Priority.WHEN_DISCOVERED,
				Priority.TIMES_DISCOVERED, Priority.ADVERTISING_SUCCESS,
				Priority.SERVICE_SEARCH_ERROR);

		List<AdvertisingDevice> activeDevicesList = deviceManager
				.getEnabledDevicesByPriority(policy);

		filterNotAdvertisableDevices(activeDevicesList);
		advertiseDevices(activeDevicesList);
	}

	private void advertiseDevices(List<AdvertisingDevice> activeDevicesList) {
		AdvertisingMediaFile currentMediaFile;

		logger.info("TO_BE_ADVERTISED {}", StringUtils
				.getSeparatedListOfAdvertisingDevices(activeDevicesList, ","));

		for (AdvertisingDevice device : activeDevicesList) {
			currentMediaFile = mediaManager.getNextMediaForDevice(device);

			if (currentMediaFile == null)
				continue;

			FilePusher filePusher = new OBEXFilePusher(device, currentMediaFile);
			filePusher.lockDeviceForTransfer();

			AdvertisingMetadataObserver observer = new AdvertisingMetadataObserver();
			filePusher.addObserver(observer);
			scheduler.submitFileTransferTask(filePusher);
		}
	}

	private void filterNotAdvertisableDevices(
			List<AdvertisingDevice> activeDevicesList) {
		List<AdvertisingDevice> toBeFilteredOut = new LinkedList<AdvertisingDevice>();

		for (AdvertisingDevice device : activeDevicesList) {
			if (filterChain.shouldFilter(device))
				toBeFilteredOut.add(device);
		}

		logger.info("NOT_ADVERTISABLE {}", StringUtils
				.getSeparatedListOfAdvertisingDevices(toBeFilteredOut, ","));

		for (AdvertisingDevice filteredOutDevice : toBeFilteredOut)
			activeDevicesList.remove(filteredOutDevice);
	}

}
