package br.ime.usp.mac499.bluetooth.common;

import java.util.concurrent.Semaphore;
import java.util.concurrent.locks.Condition;
import java.util.concurrent.locks.Lock;
import java.util.concurrent.locks.ReentrantLock;

/*
 * This file is part of MAC499 Bluetooth NFC 
 * Copyright (C) 2010 Rodolpho Atoji
 * 
 * This program is free software; you can redistribute it and/or modify it under
 * the terms of the GNU General Public License as published by the Free Software
 * Foundation; either version 2 of the License, or (at your option) any later
 * version.
 * 
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
 * details.
 * 
 * You should have received a copy of the GNU General Public License along with
 * this program; if not, write to the Free Software Foundation, Inc., 51
 * Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
 */

/**
 * Monitor for concurrent control in the access of Bluetooth adapters.
 * 
 * The monitor should be able to:
 * 
 * <ul>
 * 	<li>Allocate exclusive adapter for discovery process;</li>
 * 	<li>Limit the number of concurrent transfers per adapter;</li>
 * 	<li>Service search and file transfers are subject to the limit of
 * 	concurrent transfer per adapter.</li>
 * </ul>
 * 
 * @author Rodolpho Atoji
 */
public class BluetoothAdapterMonitor {

	/**
	 * Indicates if a discovery process is waiting to happen in the discovery
	 * adapter.
	 */
	private boolean discoveryPending;

	/**
	 * Limit of concurrent transfers per adapter.
	 */
	private final int maxConcurrentTransfers;

	/**
	 * Chosen adapter for the discovery process.
	 */
	private final int discoveryAdapter;

	/**
	 * Monitor lock.
	 */
	private final Lock discoveryLock;

	/**
	 * When wanting to use the discovery adapter, discovery process waits 
	 * here.
	 */
	private final Condition discoveryCondition;

	/**
	 * When wanting to use the discovery adapter, file transfers and
	 * service searches wait here.
	 */
	private final Condition transferCondition;
	
	/**
	 * Conditions per semaphore, to avoid exclusive locking of monitor.
	 */
	private final Condition[] semaphoreConditions;

	/**
	 * Semaphores to limit maximum processes per adapter.
	 */
	private final Semaphore adapterSemaphores[];

	/**
	 * Builds a monitor with all mandatory restrictions.
	 * 
	 * @param adapters Total of Bluetooth adapters in the system.
	 * @param discoveryAdapter Chosen adapter for discovery process.
	 * @param maxConcurrentTransfers Maximum of concurrent transfer per
	 * 	adapter.
	 */
	public BluetoothAdapterMonitor(int adapters, int discoveryAdapter,
			int maxConcurrentTransfers) {
		this.adapterSemaphores = new Semaphore[adapters];
		this.semaphoreConditions = new Condition[adapters];
		this.discoveryAdapter = discoveryAdapter;
		this.maxConcurrentTransfers = maxConcurrentTransfers;

		this.discoveryPending = false;

		this.discoveryLock = new ReentrantLock(true);
		this.discoveryCondition = discoveryLock.newCondition();
		this.transferCondition = discoveryLock.newCondition();

		for (int i = 0; i < adapterSemaphores.length; i++) {
			adapterSemaphores[i] = new Semaphore(maxConcurrentTransfers, true);
			semaphoreConditions[i] = discoveryLock.newCondition();
		}
	}

	/**
	 * ONLY discovery process should call this.
	 * 
	 * When called, monitor gives exclusive access to the specified discovery
	 * adapter. If there are devices using the discovery adapter, monitor
	 * locks the adapter until the running tasks are over, then releases it to 
	 * the discovery process.
	 */
	public void acquireAdapterForDiscovery() {
		try {
			discoveryLock.lock();
			this.discoveryPending = true;

			/* If there are other processes using, then wait */
			Semaphore discoverySemaphore = adapterSemaphores[discoveryAdapter];
			while (discoverySemaphore.availablePermits() != maxConcurrentTransfers)
				discoveryCondition.await();

			/* Get exclusive access */
			int drainedPermits = discoverySemaphore.drainPermits();
			if (drainedPermits != maxConcurrentTransfers)
				throw new RuntimeException("Monitor inconsistency detected");
		} catch (InterruptedException e) {
			e.printStackTrace();
		} finally {
			discoveryLock.unlock();
		}
	}

	/**
	 * ONLY discovery process should call this.
	 * 
	 * Releases the discovery adapter, after performing a discovery process.
	 * Any other task can request access to the adapter through 
	 * <code>acquireAdapter()</code>.
	 */
	public void releaseAdapterForDiscovery() {
		try {
			discoveryLock.lock();

			/* Release all permits */
			Semaphore discoverySemaphore = adapterSemaphores[discoveryAdapter];
			discoverySemaphore.release(maxConcurrentTransfers);

			this.discoveryPending = false;
			transferCondition.signalAll();
		} finally {
			discoveryLock.unlock();
		}
	}

	/**
	 * Request permit to use assigned adapter.
	 * 
	 * @param adapterNumber Index of one of the Bluetooth adapters.
	 */
	public void acquireAdapter(int adapterNumber) {
		try {
			discoveryLock.lock();

			/* If requested discovery adapter and discovery is running, wait */
			if (adapterNumber == discoveryAdapter) {
				while (discoveryPending)
					transferCondition.await();
			}

			/* Do not acquire drained semaphore to avoid locking monitor */
			while (adapterSemaphores[adapterNumber].availablePermits() == 0)
				semaphoreConditions[adapterNumber].await();
			
			adapterSemaphores[adapterNumber].acquire();
		} catch (InterruptedException e) {
			e.printStackTrace();
		} finally {
			discoveryLock.unlock();
		}
	}

	/**
	 * Release permit of the used adapter.
	 * 
	 * @param adapterNumber Index of one of the Bluetooth adapters.
	 */
	public void releaseAdapter(int adapterNumber) {
		try {
			discoveryLock.lock();

			adapterSemaphores[adapterNumber].release();
			semaphoreConditions[adapterNumber].signalAll();
			discoveryCondition.signalAll();
		} finally {
			discoveryLock.unlock();
		}
	}

}
