package br.ime.usp.mac499.bluetooth.common;

import java.util.concurrent.ExecutorService;
import java.util.concurrent.Executors;
import java.util.concurrent.ScheduledExecutorService;
import java.util.concurrent.TimeUnit;

import br.ime.usp.mac499.bluetooth.common.config.BluetoothParameters;
import br.ime.usp.mac499.bluetooth.common.discovery.DeviceDiscoverer;
import br.ime.usp.mac499.bluetooth.common.discovery.DiscoveryType;
import br.ime.usp.mac499.bluetooth.common.discovery.RunnableDeviceDiscoverer;
import br.ime.usp.mac499.bluetooth.common.sdp.RunnableServiceDiscoverer;
import br.ime.usp.mac499.bluetooth.common.sdp.ServiceDiscoverer;
import br.ime.usp.mac499.bluetooth.common.stack.LocalAdaptersManager;
import br.ime.usp.mac499.bluetooth.io.FilePusher;
import br.ime.usp.mac499.bluetooth.io.RunnableFilePusher;

/*
 * This file is part of MAC499 Bluetooth NFC Copyright (C) 2010 Rodolpho Atoji
 * 
 * This program is free software; you can redistribute it and/or modify it under
 * the terms of the GNU General Public License as published by the Free Software
 * Foundation; either version 2 of the License, or (at your option) any later
 * version.
 * 
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
 * details.
 * 
 * You should have received a copy of the GNU General Public License along with
 * this program; if not, write to the Free Software Foundation, Inc., 51
 * Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
 */

/**
 * Scheduler implementing a FIFO policy for scheduling the Bleutooth adapters.
 * 
 * This implementation uses fixed thread pools to assign threads to
 * submitted tasks.
 * 
 * @author Rodolpho Atoji
 */
public class GenericBluetoothAdapterScheduler implements
		BluetoothAdapterScheduler {

	private final ScheduledExecutorService discoveryPool;

	private final ExecutorService servicePool;

	private final ExecutorService transferPool;

	private final BluetoothParameters parameters;

	private final int totalAdapters;

	private final int concurrentTransfers;

	private int currentAdapter;

	/**
	 * Builds a scheduler with parameters set on: Bluetooth runtime parameters 
	 * and local adapters manager.
	 */
	public GenericBluetoothAdapterScheduler() {
		LocalAdaptersManager adaptersManager = BluetoothResources
				.getLocalAdaptersManager();
		this.parameters = BluetoothResources.getParameters();
		this.currentAdapter = 0;

		this.concurrentTransfers = parameters.getMaxConcurrentTransfers();
		this.totalAdapters = adaptersManager.getTotalAdapters();

		this.discoveryPool = Executors.newSingleThreadScheduledExecutor();
		this.servicePool = Executors.newFixedThreadPool(concurrentTransfers
				* totalAdapters);
		this.transferPool = Executors.newFixedThreadPool(concurrentTransfers
				* totalAdapters);

	}

	@Override
	public void registerDiscoveryTask(DeviceDiscoverer discovererTask,
			DiscoveryType discoveryType) {
		RunnableDeviceDiscoverer runnableDiscoverer = new RunnableDeviceDiscoverer(
				discovererTask, discoveryType);
		long discoveryInterval = parameters.getDiscoveryInterval();

		try {
			discoveryPool.scheduleAtFixedRate(runnableDiscoverer, 0,
					discoveryInterval, TimeUnit.SECONDS);
		} catch (Exception e) {
			e.printStackTrace();
		}
	}

	@Override
	public void submitServiceDiscoveryTask(ServiceDiscoverer discovererTask) {
		int nextAdapter = getNextTransferAdapter();
		RunnableServiceDiscoverer serviceDiscoverer = new RunnableServiceDiscoverer(
				discovererTask, nextAdapter);

		try {
			servicePool.submit(serviceDiscoverer);
		} catch (Exception e) {
			e.printStackTrace();
		}
	}

	@Override
	public void submitFileTransferTask(FilePusher pusherTask) {
		int nextAdapter = getNextTransferAdapter();
		RunnableFilePusher filePusher = new RunnableFilePusher(pusherTask,
				nextAdapter);

		try {
			transferPool.submit(filePusher);
		} catch (Exception e) {
			e.printStackTrace();
		}
	}

	private synchronized int getNextTransferAdapter() {
		this.currentAdapter = ++this.currentAdapter % totalAdapters;
		return currentAdapter;
	}

}
