package br.ime.usp.mac499.bluetooth.common.config;

import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.IOException;
import java.util.Properties;

import br.ime.usp.mac499.util.PathUtils;

/*
 * This file is part of MAC499 Bluetooth NFC Copyright (C) 2010 Rodolpho Atoji
 * 
 * This program is free software; you can redistribute it and/or modify it under
 * the terms of the GNU General Public License as published by the Free Software
 * Foundation; either version 2 of the License, or (at your option) any later
 * version.
 * 
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
 * details.
 * 
 * You should have received a copy of the GNU General Public License along with
 * this program; if not, write to the Free Software Foundation, Inc., 51
 * Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
 */

/**
 * Interface for the <code>bluetooth.properties</code> file.
 * 
 * The file is read and parsed by this class.
 * 
 * @author Rodolpho Atoji
 */
public class PropertiesBluetoothParameters implements BluetoothParameters {

	private final DefaultBluetoothParameters defaultParameters;

	private long discoveryInterval;

	private int bluetoothMaxConcurrent;

	private int bluetoothDiscoveryAdapter;

	private int advertisingMaxRetries;

	private int advertisingMaxSuccesses;

	private long advertisingInterval;

	/**
	 * Build a properties holder, first trying to load them from 
	 * <code>bluetooth.properties</code>. In case of failing, default
	 * parameters are loaded.
	 */
	public PropertiesBluetoothParameters() {
		this.defaultParameters = new DefaultBluetoothParameters();

		reloadProperties();
	}

	/**
	 * Reload properties from disk.
	 */
	public void reloadProperties() {
		Properties loadedProperties = reloadConfigFileFromDisk();
		refreshParameters(loadedProperties);
	}

	/**
	 * Refresh cached properties.
	 * 
	 * @param loadedProperties Loaded properties from disk.
	 */
	private void refreshParameters(Properties loadedProperties) {
		refreshDiscoveryInterval(loadedProperties);

		refreshBluetoothMaxConcurrent(loadedProperties);
		refreshBluetoothDiscoveryAdapter(loadedProperties);

		refreshAdvertisingMaxRetries(loadedProperties);
		refreshAdvertisingMaxSuccesses(loadedProperties);
		refreshAdvertisingInterval(loadedProperties);
	}

	private void refreshAdvertisingInterval(Properties loadedProperties) {
		try {
			String intervalString = loadedProperties
					.getProperty(ConfigKey.Advertising.INTERVAL.toString());
			long interval = Long.parseLong(intervalString);
			this.advertisingInterval = interval;
		} catch (Exception e) {
			this.advertisingInterval = defaultParameters
					.getAdvertisingInterval();
			System.err.println("[WARN] Defaulting "
					+ ConfigKey.Advertising.INTERVAL.toString() + " to "
					+ advertisingInterval);
		}
	}

	private void refreshAdvertisingMaxSuccesses(Properties loadedProperties) {
		try {
			String maxSuccessesString = loadedProperties
					.getProperty(ConfigKey.Advertising.MAX_SUCCESSES.toString());
			int maxSuccesses = Integer.parseInt(maxSuccessesString);
			this.advertisingMaxSuccesses = maxSuccesses;
		} catch (Exception e) {
			this.advertisingMaxSuccesses = defaultParameters
					.getMaxAdvertisingSuccesses();
			System.err.println("[WARN] Defaulting "
					+ ConfigKey.Advertising.MAX_SUCCESSES.toString() + " to "
					+ advertisingMaxSuccesses);
		}
	}

	private void refreshAdvertisingMaxRetries(Properties loadedProperties) {
		try {
			String maxRetriesString = loadedProperties
					.getProperty(ConfigKey.Advertising.MAX_RETRIES.toString());
			int maxRetries = Integer.parseInt(maxRetriesString);
			this.advertisingMaxRetries = maxRetries;
		} catch (Exception e) {
			this.advertisingMaxRetries = defaultParameters
					.getMaxAdvertisingRetries();
			System.err.println("[WARN] Defaulting "
					+ ConfigKey.Advertising.MAX_RETRIES.toString() + " to "
					+ advertisingMaxRetries);
		}
	}

	private void refreshBluetoothDiscoveryAdapter(Properties loadedProperties) {
		try {
			String discoveryAdapterString = loadedProperties
					.getProperty(ConfigKey.Bluetooth.DISCOVERY_ADAPTER
							.toString());
			int discoveryAdapter = Integer.parseInt(discoveryAdapterString);
			this.bluetoothDiscoveryAdapter = discoveryAdapter;
		} catch (Exception e) {
			this.bluetoothDiscoveryAdapter = defaultParameters
					.getDiscoveryAdapterNumber();
			System.err.println("[WARN] Defaulting "
					+ ConfigKey.Bluetooth.DISCOVERY_ADAPTER.toString() + " to "
					+ bluetoothDiscoveryAdapter);
		}
	}

	private void refreshBluetoothMaxConcurrent(Properties loadedProperties) {
		try {
			String maxConcurrentString = loadedProperties
					.getProperty(ConfigKey.Bluetooth.MAX_CONCURRENT.toString());
			int maxConcurrent = Integer.parseInt(maxConcurrentString);
			this.bluetoothMaxConcurrent = maxConcurrent;
		} catch (Exception e) {
			this.bluetoothMaxConcurrent = defaultParameters
					.getMaxConcurrentTransfers();
			System.err.println("[WARN] Defaulting "
					+ ConfigKey.Bluetooth.MAX_CONCURRENT.toString() + " to "
					+ bluetoothMaxConcurrent);
		}
	}

	private void refreshDiscoveryInterval(Properties loadedProperties) {
		try {
			String discoveryIntervalString = loadedProperties
					.getProperty(ConfigKey.Discovery.INTERVAL.toString());
			long interval = Long.parseLong(discoveryIntervalString);
			this.discoveryInterval = interval;
		} catch (Exception e) {
			this.discoveryInterval = defaultParameters.getDiscoveryInterval();
			System.err.println("[WARN] Defaulting "
					+ ConfigKey.Discovery.INTERVAL.toString() + " to "
					+ discoveryInterval);
		}
	}

	private Properties reloadConfigFileFromDisk() {
		String configPath = PathUtils.getConfigPath();
		Properties loadedProperties = null;

		try {
			FileInputStream configInputStream = new FileInputStream(configPath);
			loadedProperties = new Properties();
			loadedProperties.load(configInputStream);
		} catch (FileNotFoundException e) {
			System.err.println("[ERROR] Failed to locate config file: "
					+ configPath + ", using defaults");
		} catch (IOException e) {
			System.err.println("[ERROR] Failed to process config file: "
					+ configPath + ", using defaults");
		}

		return loadedProperties;
	}

	@Override
	public long getDiscoveryInterval() {
		return discoveryInterval;
	}

	@Override
	public int getMaxAdvertisingRetries() {
		return advertisingMaxRetries;
	}

	@Override
	public int getMaxAdvertisingSuccesses() {
		return advertisingMaxSuccesses;
	}

	@Override
	public int getMaxConcurrentTransfers() {
		return bluetoothMaxConcurrent;
	}

	@Override
	public int getDiscoveryAdapterNumber() {
		return bluetoothDiscoveryAdapter;
	}

	@Override
	public long getAdvertisingInterval() {
		return advertisingInterval;
	}

}
