package br.ime.usp.mac499.bluetooth.common.discovery;

import java.util.Date;
import java.util.LinkedList;
import java.util.List;
import java.util.concurrent.Semaphore;
import java.util.concurrent.locks.Condition;
import java.util.concurrent.locks.Lock;
import java.util.concurrent.locks.ReentrantLock;

import javax.bluetooth.BluetoothStateException;
import javax.bluetooth.DeviceClass;
import javax.bluetooth.DiscoveryAgent;
import javax.bluetooth.DiscoveryListener;
import javax.bluetooth.LocalDevice;
import javax.bluetooth.RemoteDevice;
import javax.bluetooth.ServiceRecord;

/*
 * This file is part of MAC499 Bluetooth NFC Copyright (C) 2010 Rodolpho Atoji
 * 
 * This program is free software; you can redistribute it and/or modify it under
 * the terms of the GNU General Public License as published by the Free Software
 * Foundation; either version 2 of the License, or (at your option) any later
 * version.
 * 
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
 * details.
 * 
 * You should have received a copy of the GNU General Public License along with
 * this program; if not, write to the Free Software Foundation, Inc., 51
 * Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
 */

/**
 * Simple synchronized device discoverer.
 * 
 * @author Rodolpho Atoji
 */
public class GenericDeviceDiscoverer implements DiscoveryListener,
		DeviceDiscoverer {

	private int inquiryStatus;

	private boolean performingInquiry;

	private final Lock inquiryLock;

	private final Condition completeCondition;

	private final Semaphore inquirySemaphore;

	private final List<DiscoveredDevice> discoveredDevices;

	private final List<DiscoveryObserver> observerList;

	private static DeviceDiscoverer instance;

	private GenericDeviceDiscoverer() {
		this.inquiryLock = new ReentrantLock();
		this.completeCondition = inquiryLock.newCondition();
		this.inquirySemaphore = new Semaphore(1);
		this.discoveredDevices = new LinkedList<DiscoveredDevice>();
		this.observerList = new LinkedList<DiscoveryObserver>();
	}

	/**
	 * @return Singleton instance of discoverer.
	 */
	public static DeviceDiscoverer getInstance() {
		synchronized (GenericDeviceDiscoverer.class) {
			if (instance == null)
				instance = new GenericDeviceDiscoverer();
			return instance;
		}
	}

	@Override
	public List<DiscoveredDevice> doSyncInquiry(DiscoveryType discoveryType)
			throws BluetoothStateException {
		try {
			inquirySemaphore.acquire();
		} catch (InterruptedException e) {
			throw new BluetoothStateException(
					"Already performing device discovery");
		}

		inquiryLock.lock();

		try {
			performingInquiry = true;
			discoveredDevices.clear();

			LocalDevice localDevice = LocalDevice.getLocalDevice();
			DiscoveryAgent discoveryAgent = localDevice.getDiscoveryAgent();

			Date startTime = new Date();
			discoveryAgent.startInquiry(discoveryType.getInquiryCode(), this);
			while (performingInquiry)
				completeCondition.await();
			Date endTime = new Date();

			notifyObservers(discoveredDevices, startTime, endTime);

			return discoveredDevices;
		} catch (InterruptedException e) {
			return discoveredDevices;
		} finally {
			inquirySemaphore.release();
			inquiryLock.unlock();
		}
	}

	@Override
	public void deviceDiscovered(RemoteDevice remoteDevice,
			DeviceClass deviceClass) {
		DiscoveredDevice discoveredDevice = new DiscoveredDevice(remoteDevice,
				deviceClass, new Date());
		discoveredDevices.add(discoveredDevice);
	}

	@Override
	public void inquiryCompleted(int status) {
		inquiryLock.lock();

		try {
			this.inquiryStatus = status;
			performingInquiry = false;
			completeCondition.signalAll();
		} finally {
			inquiryLock.unlock();
		}
	}

	@Override
	public void addDiscoveryObserver(DiscoveryObserver observer) {
		this.observerList.add(observer);
	}

	@Override
	public void removeDiscoveryObserver(DiscoveryObserver observer) {
		this.observerList.remove(observer);
	}

	private void notifyObservers(List<DiscoveredDevice> discoveredList,
			Date startTime, Date endTime) {
		DiscoveryStatus status = getLastStatus();

		for (DiscoveryObserver observer : observerList)
			observer.notifyDevicesDiscovered(discoveredList, startTime,
					endTime, status);
	}

	@Override
	public DiscoveryStatus getLastStatus() {
		return DiscoveryStatus.getStatusMapping(inquiryStatus);
	}

	@Override
	public void serviceSearchCompleted(int arg0, int arg1) {
		throw new UnsupportedOperationException(
				"This method must be called on ServiceDiscoverer");
	}

	@Override
	public void servicesDiscovered(int arg0, ServiceRecord[] arg1) {
		throw new UnsupportedOperationException(
				"This method must be called on ServiceDiscoverer");
	}

}
