package br.ime.usp.mac499.bluetooth.common.discovery;

import java.util.LinkedList;
import java.util.List;

import javax.bluetooth.BluetoothStateException;

import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import br.ime.usp.mac499.bluetooth.common.BluetoothAdapterMonitor;
import br.ime.usp.mac499.bluetooth.common.BluetoothResources;
import br.ime.usp.mac499.bluetooth.common.stack.LocalAdapterWrapper;
import br.ime.usp.mac499.bluetooth.common.stack.LocalAdaptersManager;
import br.ime.usp.mac499.util.StringUtils;

/*
 * This file is part of MAC499 Bluetooth NFC Copyright (C) 2010 Rodolpho Atoji
 * 
 * This program is free software; you can redistribute it and/or modify it under
 * the terms of the GNU General Public License as published by the Free Software
 * Foundation; either version 2 of the License, or (at your option) any later
 * version.
 * 
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
 * details.
 * 
 * You should have received a copy of the GNU General Public License along with
 * this program; if not, write to the Free Software Foundation, Inc., 51
 * Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
 */

/**
 * Thread-wrapper for a device discoverer.
 * 
 * @author Rodolpho Atoji
 */
public class RunnableDeviceDiscoverer implements Runnable {

	private final DeviceDiscoverer discoverer;

	private final DiscoveryType discoveryType;

	private final Logger logger;

	/**
	 * Builds a thread-wrapper for a discoverer with the specified
	 * discovery type. 
	 * 
	 * @param discoverer Device discoverer.
	 * @param discoveryType Discovery type.
	 */
	public RunnableDeviceDiscoverer(DeviceDiscoverer discoverer,
			DiscoveryType discoveryType) {
		this.discoverer = discoverer;
		this.discoveryType = discoveryType;
		this.logger = LoggerFactory.getLogger(RunnableDeviceDiscoverer.class);
	}

	@Override
	public void run() {
		long start = 0;
		List<DiscoveredDevice> devices = new LinkedList<DiscoveredDevice>();

		BluetoothAdapterMonitor monitor = BluetoothResources
				.getBluetoothAdapterMonitor();
		LocalAdaptersManager adaptersManager = BluetoothResources
				.getLocalAdaptersManager();

		/* Select natively which Bluetooth adapter to use */
		Object adapter = adaptersManager.getDiscoveryAdapter();
		LocalAdapterWrapper adapterWrapper = adaptersManager
				.getLocalAdapterWrapper();

		try {
			adapterWrapper.selectAdapterForUse(adapter);
			monitor.acquireAdapterForDiscovery();

			logger.info("INQ_START  adapter: {}", adapter);
			start = System.currentTimeMillis();

			devices = discoverer.doSyncInquiry(discoveryType);
		} catch (BluetoothStateException e) {
			e.printStackTrace();
		} finally {
			DiscoveryStatus status = discoverer.getLastStatus();
			long end = System.currentTimeMillis();
			String discoveredList = StringUtils
					.getSeparatedListOfDiscoveredDevices(devices, ",");
			logger.info("DISCOVERED {}", discoveredList);
			logger.info("INQ_END    status: {} discovered: {} duration: {}",
					new Object[] { status, devices.size(), (end - start) });

			adapterWrapper.releaseUsedAdapter();
			monitor.releaseAdapterForDiscovery();
		}
	}

}