package br.ime.usp.mac499.bluetooth.common.sdp;

import java.util.Date;
import java.util.LinkedList;
import java.util.concurrent.locks.Condition;
import java.util.concurrent.locks.Lock;
import java.util.concurrent.locks.ReentrantLock;

import javax.bluetooth.BluetoothStateException;
import javax.bluetooth.DeviceClass;
import javax.bluetooth.DiscoveryListener;
import javax.bluetooth.LocalDevice;
import javax.bluetooth.RemoteDevice;
import javax.bluetooth.ServiceRecord;
import javax.bluetooth.UUID;

/*
 * This file is part of MAC499 Bluetooth NFC Copyright (C) 2010 Rodolpho Atoji
 * 
 * This program is free software; you can redistribute it and/or modify it under
 * the terms of the GNU General Public License as published by the Free Software
 * Foundation; either version 2 of the License, or (at your option) any later
 * version.
 * 
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
 * details.
 * 
 * You should have received a copy of the GNU General Public License along with
 * this program; if not, write to the Free Software Foundation, Inc., 51
 * Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
 */

/**
 * Generic implementation of a service discovery task.
 * 
 * @author Rodolpho Atoji
 */
public class GenericServiceDiscoverer implements DiscoveryListener,
		ServiceDiscoverer {

	private final int[] attrSet;

	private final UUID[] uuidSet;

	private final RemoteDevice btDev;

	private final Lock searchLock;

	private final Condition searchingCondition;

	private boolean searchPerformed;

	private ServiceStatus sdpResult;

	private ServiceRecord[] serviceRecords;

	private final LinkedList<ServiceSearchObserver> observerList;

	/**
	 * Build a generic discoverer for the JSR 82 parameters.
	 * 
	 * @param attrSet Attributes to search.
	 * @param uuidSet UUIDs to search.
	 * @param btDev Bluetooth device.
	 */
	public GenericServiceDiscoverer(int[] attrSet, UUID[] uuidSet,
			RemoteDevice btDev) {
		this.attrSet = attrSet;
		this.uuidSet = uuidSet;
		this.btDev = btDev;
		this.sdpResult = ServiceStatus.UNKNOWN;

		this.observerList = new LinkedList<ServiceSearchObserver>();

		this.searchPerformed = false;
		this.searchLock = new ReentrantLock();
		this.searchingCondition = searchLock.newCondition();
	}

	@Override
	public void doServiceSearch() throws BluetoothStateException {
		LocalDevice localDevice = LocalDevice.getLocalDevice();

		searchLock.lock();
		try {
			Date startTime = new Date();
			localDevice.getDiscoveryAgent().searchServices(attrSet, uuidSet,
					btDev, this);
			while (!searchPerformed)
				searchingCondition.await();
			Date endTime = new Date();

			notifyObservers(startTime, endTime);
		} catch (InterruptedException e) {
			e.printStackTrace();
		} finally {
			searchLock.unlock();
		}
	}

	@Override
	public void servicesDiscovered(int transID, ServiceRecord[] servRecord) {
		this.serviceRecords = servRecord;
	}

	@Override
	public void serviceSearchCompleted(int transID, int respCode) {
		searchLock.lock();

		try {
			switch (respCode) {
			case DiscoveryListener.SERVICE_SEARCH_COMPLETED:
				sdpResult = ServiceStatus.COMPLETED;
				break;
			case DiscoveryListener.SERVICE_SEARCH_DEVICE_NOT_REACHABLE:
				sdpResult = ServiceStatus.NOT_REACHABLE;
				break;
			case DiscoveryListener.SERVICE_SEARCH_ERROR:
				sdpResult = ServiceStatus.ERROR;
				break;
			case DiscoveryListener.SERVICE_SEARCH_NO_RECORDS:
				sdpResult = ServiceStatus.NO_RECORDS;
				break;
			case DiscoveryListener.SERVICE_SEARCH_TERMINATED:
				sdpResult = ServiceStatus.TERMINATED;
				break;
			default:
				sdpResult = ServiceStatus.UNKNOWN;
			}
		} finally {
			searchPerformed = true;
			searchingCondition.signalAll();
			searchLock.unlock();
		}
	}

	@Override
	public ServiceStatus getSdpResult() {
		return sdpResult;
	}

	@Override
	public ServiceRecord[] getServiceRecords() {
		return serviceRecords;
	}

	@Override
	public void deviceDiscovered(RemoteDevice btDevice, DeviceClass cod) {
		throw new UnsupportedOperationException(
				"This method must be called on DeviceDiscoverer");
	}

	@Override
	public void inquiryCompleted(int discType) {
		throw new UnsupportedOperationException(
				"This method must be called on DeviceDiscoverer");
	}

	@Override
	public RemoteDevice getRemoteDevice() {
		return btDev;
	}

	@Override
	public void addServiceSearchObserver(ServiceSearchObserver observer) {
		this.observerList.add(observer);
	}

	@Override
	public void removeServiceSearchObserver(ServiceSearchObserver observer) {
		this.observerList.remove(observer);
	}

	private void notifyObservers(Date startTime, Date endTime) {
		ServiceSearchAttempt attempt = new ServiceSearchAttempt(sdpResult,
				btDev, serviceRecords, startTime, endTime);

		for (ServiceSearchObserver observer : observerList)
			observer.notifySearchResult(attempt);
	}

}
