package br.ime.usp.mac499.bluetooth.common.sdp;

import java.util.Date;
import java.util.List;

import javax.bluetooth.RemoteDevice;

import br.ime.usp.mac499.bluetooth.common.BluetoothAdapterScheduler;
import br.ime.usp.mac499.bluetooth.common.discovery.DiscoveredDevice;
import br.ime.usp.mac499.bluetooth.common.discovery.DiscoveryObserver;
import br.ime.usp.mac499.bluetooth.common.discovery.DiscoveryStatus;
import br.ime.usp.mac499.bluetooth.db.ServiceSearchMetadataObserver;
import br.ime.usp.mac499.bluetooth.model.AdvertisingDevice;
import br.ime.usp.mac499.bluetooth.model.AdvertisingDeviceManager;

/*
 * This file is part of MAC499 Bluetooth NFC 
 * Copyright (C) 2010 Rodolpho Atoji
 * 
 * This program is free software; you can redistribute it and/or modify it under
 * the terms of the GNU General Public License as published by the Free Software
 * Foundation; either version 2 of the License, or (at your option) any later
 * version.
 * 
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
 * details.
 * 
 * You should have received a copy of the GNU General Public License along with
 * this program; if not, write to the Free Software Foundation, Inc., 51
 * Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
 */

/**
 * Observer of discovery process which dispatches service searches for
 * every discovered device which has not been searched successfully in
 * previous attempts.
 * 
 * @author Rodolpho Atoji
 */
public class ServiceSearchDispatcher implements DiscoveryObserver {

	private final BluetoothAdapterScheduler scheduler;

	private final AdvertisingDeviceManager deviceManager;

	/**
	 * Constructs a dispatcher which updates the advertising device manager
	 * and submits a service search task to the scheduler.
	 * 
	 * @param deviceManager Advertising device manager.
	 * @param scheduler Bluetooth adapter scheduler.
	 */
	public ServiceSearchDispatcher(AdvertisingDeviceManager deviceManager,
			BluetoothAdapterScheduler scheduler) {
		this.deviceManager = deviceManager;
		this.scheduler = scheduler;
	}

	@Override
	public void notifyDevicesDiscovered(List<DiscoveredDevice> discoveredList,
			Date startTime, Date endTime, DiscoveryStatus status) {
		ServiceStatus serviceSearchStatus;
		RemoteDevice currentRemoteDevice;
		String currentMacAddress;

		for (DiscoveredDevice discoveredDevice : discoveredList) {
			currentRemoteDevice = discoveredDevice.getRemoteDevice();
			currentMacAddress = currentRemoteDevice.getBluetoothAddress();
			serviceSearchStatus = deviceManager
					.getServiceSearchStatus(currentMacAddress);

			if (deviceManager.isDeviceEnabled(currentMacAddress)) {
				if (shouldPerformServiceSearch(serviceSearchStatus))
					dispatchServiceSearch(currentMacAddress);
			}
		}
	}

	private void dispatchServiceSearch(String currentMacAddress) {
		AdvertisingDevice device = deviceManager.getDevice(currentMacAddress);
		device.setServiceSearchStatus(ServiceStatus.IN_PROGRESS);

		ServiceDiscoverer serviceDiscoverer = ServiceDiscovererFactory
				.buildSingleServiceDiscoverer(device, ServiceType.OBEX_PUSH);

		ServiceSearchObserver searchObserver = new ServiceSearchProcessor(
				deviceManager);
		serviceDiscoverer.addServiceSearchObserver(searchObserver);

		ServiceSearchMetadataObserver metadataObserver = new ServiceSearchMetadataObserver();
		serviceDiscoverer.addServiceSearchObserver(metadataObserver);

		scheduler.submitServiceDiscoveryTask(serviceDiscoverer);
	}

	/**
	 * A service search should not be performed if it was already
	 * successfully performed previously, if it is currently in progress or
	 * the device does not supports service search.
	 * 
	 * @param serviceSearchStatus Status of a previous service search.
	 * @return <code>true</code> if the conditions above are met.
	 */
	private boolean shouldPerformServiceSearch(ServiceStatus serviceSearchStatus) {
		if ((serviceSearchStatus == ServiceStatus.COMPLETED)
				|| (serviceSearchStatus == ServiceStatus.IN_PROGRESS)
				|| (serviceSearchStatus == ServiceStatus.INVALID))
			return false;
		return true;
	}
}
