package br.ime.usp.mac499.bluetooth.common.stack;

import java.text.DecimalFormat;
import java.text.NumberFormat;
import java.util.ArrayList;

import javax.bluetooth.BluetoothStateException;
import javax.bluetooth.DiscoveryAgent;
import javax.bluetooth.LocalDevice;

import com.intel.bluetooth.BlueCoveConfigProperties;
import com.intel.bluetooth.BlueCoveImpl;

/*
 * This file is part of MAC499 Bluetooth NFC Copyright (C) 2010 Rodolpho Atoji
 * 
 * This program is free software; you can redistribute it and/or modify it under
 * the terms of the GNU General Public License as published by the Free Software
 * Foundation; either version 2 of the License, or (at your option) any later
 * version.
 * 
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
 * details.
 * 
 * You should have received a copy of the GNU General Public License along with
 * this program; if not, write to the Free Software Foundation, Inc., 51
 * Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
 */

/**
 * Bluecove Emulator local Bluetooth adapters manager.
 * 
 * This class is responsible for instantiating a number of emulated
 * Bluetooth transmitters.
 * 
 * @author Rodolpho Atoji
 */
public class BluecoveEmulatorLocalAdaptersManager implements
		LocalAdaptersManager {

	public final static int DEFAULT_DISCOVERY_ADAPTER = 0;
	
	private final int maxAdapters;

	private final BluecoveLocalAdapterWrapper wrapper = new BluecoveLocalAdapterWrapper();

	private Object[] adapters;
	
	/**
	 * Builds a local adapters manager with the specified amount of emulated
	 * adapters.
	 * 
	 * @param maxAdapters Number of adapters to instantiate.
	 */
	public BluecoveEmulatorLocalAdaptersManager(int maxAdapters) {
		this.maxAdapters = maxAdapters;
	}

	@Override
	public void initialize() throws BluetoothStateException {
		System.out.println("[EMULATOR] Initializing local adapters...");

		System.setProperty("bluecove.stack", "emulator");
		BlueCoveImpl.setConfigProperty(BlueCoveConfigProperties.PROPERTY_STACK,
				BlueCoveImpl.STACK_EMULATOR);
		BlueCoveImpl.setConfigProperty(
				BlueCoveConfigProperties.PROPERTY_EMULATOR_PORT, "0");

		String adapterSuffix;
		Object currentAdapter;
		ArrayList<Object> initializedAdapters = new ArrayList<Object>();
		
		for (int i = 0; i < maxAdapters; i++) {
			adapterSuffix = getSuffix(i);
			try {
				BlueCoveImpl.setConfigProperty(	
						BlueCoveConfigProperties.PROPERTY_LOCAL_DEVICE_ADDRESS,
						"FF" + adapterSuffix);
				BlueCoveImpl.useThreadLocalBluetoothStack();
				
				LocalDevice.getLocalDevice().setDiscoverable(
						DiscoveryAgent.NOT_DISCOVERABLE);

				currentAdapter = BlueCoveImpl.getThreadBluetoothStackID();
				System.out.println("[EMULATOR] " + currentAdapter);

				initializedAdapters.add(currentAdapter);
			} finally {
				BlueCoveImpl.releaseThreadBluetoothStack();
			}
		}

		adapters = initializedAdapters.toArray();
		BlueCoveImpl.setThreadBluetoothStackID(null);
	}

	private String getSuffix(int i) {
		NumberFormat numberFormat = new DecimalFormat("0000000000");		
		return numberFormat.format(i);
	}

	@Override
	public Object getDiscoveryAdapter() {
		if ((adapters != null) && (adapters.length > 0))
			return adapters[DEFAULT_DISCOVERY_ADAPTER];
		return null;
	}

	@Override
	public int getTotalAdapters() {
		if (adapters != null)
			return adapters.length;
		return 0;
	}

	@Override
	public Object getAdapter(int adapterNumber) {
		if (adapters != null) {
			if ((adapterNumber >= 0) && (adapterNumber < adapters.length))
				return adapters[adapterNumber];
		}
		return null;
	}

	@Override
	public LocalAdapterWrapper getLocalAdapterWrapper() {
		return wrapper;
	}

}
