package br.ime.usp.mac499.bluetooth.io;

import java.io.BufferedOutputStream;
import java.io.IOException;
import java.util.Date;
import java.util.LinkedList;
import java.util.List;

import javax.microedition.io.Connection;
import javax.microedition.io.Connector;
import javax.obex.ClientSession;
import javax.obex.HeaderSet;
import javax.obex.Operation;

import br.ime.usp.mac499.bluetooth.db.AdvertisingMetadataObserver;
import br.ime.usp.mac499.bluetooth.model.AdvertisementAttempt;
import br.ime.usp.mac499.bluetooth.model.AdvertisingDevice;
import br.ime.usp.mac499.bluetooth.model.AdvertisingMediaFile;

/*
 * This file is part of MAC499 Bluetooth NFC Copyright (C) 2010 Rodolpho Atoji
 * 
 * This program is free software; you can redistribute it and/or modify it under
 * the terms of the GNU General Public License as published by the Free Software
 * Foundation; either version 2 of the License, or (at your option) any later
 * version.
 * 
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
 * details.
 * 
 * You should have received a copy of the GNU General Public License along with
 * this program; if not, write to the Free Software Foundation, Inc., 51
 * Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
 */

/**
 * Bluetooth-specific OBEX (OBject EXchange) file pusher.
 * 
 * @author Rodolpho Atoji
 */
public class OBEXFilePusher implements FilePusher {

	private final AdvertisingDevice device;

	private final AdvertisingMediaFile mediaFile;

	private OBEXPushStatus pushStatus;
	
	private final List<AdvertisingMetadataObserver> observerList;

	/**
	 * Builds a pusher of mediaFile for the target device.
	 * 
	 * @param device Target device.
	 * @param mediaFile Multimedia content.
	 */
	public OBEXFilePusher(AdvertisingDevice device,
			AdvertisingMediaFile mediaFile) {
		this.device = device;
		this.mediaFile = mediaFile;
		this.pushStatus = OBEXPushStatus.UNKNOWN;
		this.observerList = new LinkedList<AdvertisingMetadataObserver>();
	}

	@Override
	public void pushFile() throws IOException {
		String connectionURL = device.getConnectionURL();

		if (connectionURL == null)
			throw new IOException("Cannot get connection URL");

		Date startTime = new Date();
		Date endTime = new Date();

		Connection connection = null;
		
		try {
			startTime = new Date();

			connection = Connector.open(connectionURL);
			ClientSession clientSession = (ClientSession) connection;
			HeaderSet headerSet = clientSession.createHeaderSet();

			clientSession.connect(headerSet);

			headerSet.setHeader(HeaderSet.NAME, mediaFile.getName());
			headerSet.setHeader(HeaderSet.TYPE, mediaFile.getMimeType());
			headerSet.setHeader(HeaderSet.LENGTH, mediaFile.getContentSize());

			Operation putOperation = clientSession.put(headerSet);

			pushStatus = OBEXPushStatus.IN_PROGRESS;
			BufferedOutputStream outputStream = new BufferedOutputStream(
					putOperation.openOutputStream());
			outputStream.write(mediaFile.getContent());

			outputStream.close();
			putOperation.close();
			clientSession.disconnect(null);
			connection.close();

			pushStatus = OBEXPushStatus.SUCCESS;
			endTime = new Date();
		} catch (IOException e) {
			if (connection != null)
				connection.close();
			
			String stackErrorMessage = e.getMessage();
			pushStatus = OBEXPushStatus.getErrorMapping(stackErrorMessage);
		} finally {
			AdvertisementAttempt attempt = new AdvertisementAttempt(startTime,
					endTime, mediaFile, pushStatus);
			device.addAdvertisementAttempt(attempt);
			notifyObservers(attempt, device, mediaFile);
		}
	}

	private void notifyObservers(AdvertisementAttempt attempt,
			AdvertisingDevice device, AdvertisingMediaFile mediaFile) {
		for (AdvertisingMetadataObserver observer : observerList)
			observer.notifyObserver(attempt, device, mediaFile);
	}

	@Override
	public OBEXPushStatus getPushStatus() {
		return pushStatus;
	}

	@Override
	public String getMacAddressDestination() {
		return device.getMacAddress();
	}

	@Override
	public String getFilename() {
		return mediaFile.getName();
	}

	@Override
	public boolean lockDeviceForTransfer() {
		return device.lockForCommunication(true);
	}

	@Override
	public boolean unlockDeviceForTransfer() {
		return device.lockForCommunication(false);
	}

	@Override
	public void addObserver(AdvertisingMetadataObserver observer) {
		observerList.add(observer);
	}

}
