package br.ime.usp.mac499.bluetooth.model;

import java.util.Collections;
import java.util.Date;
import java.util.LinkedList;
import java.util.List;
import java.util.concurrent.atomic.AtomicBoolean;

import javax.bluetooth.DeviceClass;
import javax.bluetooth.RemoteDevice;
import javax.bluetooth.ServiceRecord;

import br.ime.usp.mac499.bluetooth.common.sdp.ServiceStatus;
import br.ime.usp.mac499.bluetooth.io.OBEXPushStatus;

/*
 * This file is part of MAC499 Bluetooth NFC Copyright (C) 2010 Rodolpho Atoji
 * 
 * This program is free software; you can redistribute it and/or modify it under
 * the terms of the GNU General Public License as published by the Free Software
 * Foundation; either version 2 of the License, or (at your option) any later
 * version.
 * 
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
 * details.
 * 
 * You should have received a copy of the GNU General Public License along with
 * this program; if not, write to the Free Software Foundation, Inc., 51
 * Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
 */

/**
 * Representation of a Bluetooth-enabled target device.
 * 
 * @author Rodolpho Atoji
 */
public class AdvertisingDevice {

	/**
	 * JSR 82 metadata about target device.
	 */
	private final RemoteDevice remoteDevice;

	/**
	 * JST 82 device class.
	 */
	private final DeviceClass deviceClass;

	/**
	 * Service record about the services supported by the device.
	 */
	private ServiceRecord obexServiceRecord;

	/**
	 * Status of the last service search performed on the device.
	 */
	private ServiceStatus serviceSearchStatus;

	/**
	 * Device is frequently found on device discovery?
	 */
	private boolean enabled;
	
	/**
	 * Device supports NFC?
	 */
	private boolean nfcInitiated;

	/**
	 * Lock for communication with device.
	 */
	private final AtomicBoolean deviceLocked;

	/**
	 * Updated timestamp for each discovery process.
	 */
	private Date lastDiscoveredTimestamp;

	/**
	 * Number of times the device is discovered.
	 */
	private long timesDiscovered;

	/**
	 * Number of failed advertisement attempts.
	 */
	private int failedAdvertisementAttempts;

	/**
	 * All advertisement attempts made to this device.
	 */
	private List<AdvertisementAttempt> attemptsList;

	/**
	 * Bluetooth MAC address.
	 */
	private String macAddress;

	/**
	 * JSR 82 connection URL.
	 */
	private String connectionURL;

	/**
	 * Number of service search failures.
	 */
	private int serviceSearchFailures;

	/**
	 * To avoid overflow, how many discoveries of the device will be counted.
	 */
	public static final long TIMES_DISCOVERED_LIMIT = 100;

	/**
	 * Constructor for devices found by the usual discovery process.
	 * 
	 * @param remoteDevice JSR 82 remote device metadata.
	 * @param deviceClass JSR 82 device class.
	 * @param timestamp Time of discovery.
	 */
	public AdvertisingDevice(RemoteDevice remoteDevice,
			DeviceClass deviceClass, Date timestamp) {
		this.remoteDevice = remoteDevice;
		this.deviceClass = deviceClass;
		this.lastDiscoveredTimestamp = timestamp;
		this.enabled = true;
		this.serviceSearchStatus = ServiceStatus.UNKNOWN;
		this.deviceLocked = new AtomicBoolean(false);
		this.timesDiscovered = 0;
		this.serviceSearchFailures = 0;
		this.failedAdvertisementAttempts = 0;

		this.attemptsList = Collections
				.synchronizedList(new LinkedList<AdvertisementAttempt>());
	}

	/**
	 * Constructor for devices found by a NFC touch.
	 * 
	 * @param macAddress Bluetooth MAC address.
	 * @param btConnectionURL JSR 82 connection URL.
	 * @param timestamp Touch time.
	 */
	public AdvertisingDevice(String macAddress, String btConnectionURL,
			Date timestamp) {
		this.remoteDevice = null;
		this.deviceClass = null;

		this.macAddress = macAddress;
		this.connectionURL = btConnectionURL;

		this.lastDiscoveredTimestamp = timestamp;
		this.enabled = true;
		this.serviceSearchStatus = ServiceStatus.COMPLETED;
		this.deviceLocked = new AtomicBoolean(false);
		this.timesDiscovered = 0;
		this.serviceSearchFailures = 0;
		this.failedAdvertisementAttempts = 0;

		this.attemptsList = Collections
				.synchronizedList(new LinkedList<AdvertisementAttempt>());
	}

	/**
	 * @return JSR 82 connection URL.
	 */
	public String getConnectionURL() {
		if (obexServiceRecord != null)
			return obexServiceRecord.getConnectionURL(
					ServiceRecord.NOAUTHENTICATE_NOENCRYPT, false);
		return connectionURL;
	}

	/**
	 * Set JSR 82 connection URL.
	 * 
	 * @param connectionURL JSR 82 connection URL.
	 */
	public void setConnectionURL(String connectionURL) {
		this.connectionURL = connectionURL;
	}

	public String getMacAddress() {
		if (remoteDevice != null)
			return remoteDevice.getBluetoothAddress();
		if (macAddress != null)
			return macAddress;
		return null;
	}

	public int getMajorDeviceClass() {
		return deviceClass.getMajorDeviceClass();
	}

	public RemoteDevice getRemoteDevice() {
		return remoteDevice;
	}

	public void setObexServiceRecord(ServiceRecord obexServiceRecord) {
		this.obexServiceRecord = obexServiceRecord;
	}

	public DeviceClass getDeviceClass() {
		return deviceClass;
	}

	public ServiceStatus getServiceSearchStatus() {
		return serviceSearchStatus;
	}

	public void setServiceSearchStatus(ServiceStatus serviceSearchStatus) {
		this.serviceSearchStatus = serviceSearchStatus;
	}

	/**
	 * Check if device has been detected frequently in the last discovery
	 * processes.
	 * 
	 * @return <code>true</code> if device has been detected in the last
	 * discovery processes.
	 */
	public boolean isEnabled() {
		return enabled;
	}

	/**
	 * Updates device's current discovery detection.
	 * 
	 * @param enabled <code>false</code> if device has stopped being
	 * detected in the discovery process.
	 */
	public void setEnabled(boolean enabled) {
		this.enabled = enabled;
	}

	public void addAdvertisementAttempt(AdvertisementAttempt attempt) {
		synchronized (attemptsList) {
			attemptsList.add(attempt);

			if ((attempt.getPushStatus() != OBEXPushStatus.SUCCESS)
					&& (attempt.getPushStatus() != OBEXPushStatus.IN_PROGRESS))
				this.failedAdvertisementAttempts++;
		}
	}

	public int getTotalAdvertisementAttempts() {
		synchronized (attemptsList) {
			return attemptsList.size();
		}
	}

	public int getSuccesfulAdvertisementAttempts() {
		synchronized (attemptsList) {
			return (attemptsList.size() - failedAdvertisementAttempts);
		}
	}

	public int getFailedAdvertisementAttempts() {
		return failedAdvertisementAttempts;
	}

	public Date getLastDiscoveredTimestamp() {
		return lastDiscoveredTimestamp;
	}

	public void setLastDiscoveredTimestamp(Date lastDiscoveredTimestamp) {
		this.lastDiscoveredTimestamp = lastDiscoveredTimestamp;
	}

	/**
	 * Try to get exclusive access to device.
	 * 
	 * @param lockIntent <code>true</code> if trying to lock, 
	 * <code>false</code> if trying to unlock.
	 * @return <code>true</code> if successfully locked, <code>false</code>
	 * otherwise.
	 */
	public boolean lockForCommunication(boolean lockIntent) {
		boolean canLock = deviceLocked.compareAndSet(!lockIntent, lockIntent);
		return canLock;
	}
	
	public boolean isLockedForCommunication() {
		return deviceLocked.get();
	}

	public void incrementTimesDiscovered() {
		if (timesDiscovered == TIMES_DISCOVERED_LIMIT)
			return;
		this.timesDiscovered++;
	}

	public long getTimesDiscovered() {
		return timesDiscovered;
	}

	public long getServiceSearchFailures() {
		return this.serviceSearchFailures;
	}

	public void incrementServiceSearchFailures() {
		this.serviceSearchFailures++;
	}

	/**
	 * Check if some multimedia content has already been accepted by
	 * this device.
	 *  
	 * @param mediaFile Media content to check.
	 * @return <code>true</code> if already accepted, <code>false</code>
	 * otherwise.
	 */
	public boolean hasAlreadyBeenAdvertisedWith(AdvertisingMediaFile mediaFile) {
		AdvertisingMediaFile currentMediaFile;
		
		for (AdvertisementAttempt attempt : attemptsList) {
			if (attempt.getPushStatus() == OBEXPushStatus.SUCCESS) {
				currentMediaFile = attempt.getMediaFile();
				if (currentMediaFile.equals(mediaFile))
					return true;
			}
		}
		
		return false;
	}

	/**
	 * Mark device as NFC capable.
	 * 
	 * @param nfcInitiated <code>true</code> if NFC capable.
	 */
	public void setNfcInitiated(boolean nfcInitiated) {
		this.nfcInitiated = nfcInitiated;
	}

	/**
	 * @return <code>true</code> if device has NFC capabilities,
	 * <code>false</code> otherwise.
	 */
	public boolean isNfcInitiated() {
		return nfcInitiated;
	}

}
