package br.ime.usp.mac499.bluetooth.model;

import java.util.Collections;
import java.util.Date;
import java.util.LinkedList;
import java.util.List;
import java.util.concurrent.ConcurrentHashMap;

import javax.bluetooth.ServiceRecord;

import br.ime.usp.mac499.bluetooth.common.sdp.ServiceStatus;

/*
 * This file is part of MAC499 Bluetooth NFC Copyright (C) 2010 Rodolpho Atoji
 * 
 * This program is free software; you can redistribute it and/or modify it under
 * the terms of the GNU General Public License as published by the Free Software
 * Foundation; either version 2 of the License, or (at your option) any later
 * version.
 * 
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
 * details.
 * 
 * You should have received a copy of the GNU General Public License along with
 * this program; if not, write to the Free Software Foundation, Inc., 51
 * Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
 */

/**
 * Central manager of all discovered Bluetooth devices.
 * 
 * All device changes should preferably use the methods of this class
 * to avoid inconsistency. 
 * 
 * @author Rodolpho Atoji
 */
public class AdvertisingDeviceManager {

	private final ConcurrentHashMap<String, AdvertisingDevice> deviceMap;

	public AdvertisingDeviceManager() {
		this.deviceMap = new ConcurrentHashMap<String, AdvertisingDevice>();
	}

	/**
	 * Add a device to the manager.
	 * 
	 * Manager checks if the MAC address is not already known. If it
	 * is, the new device is discarded.
	 * 
	 * @param advertisingDevice Newly discovered advertising device.
	 */
	public void addDevice(AdvertisingDevice advertisingDevice) {
		String macAddress = advertisingDevice.getMacAddress();

		if (!deviceMap.containsKey(macAddress))
			deviceMap.put(macAddress, advertisingDevice);
	}

	/**
	 * Add results of service search on a device.
	 * 
	 * @param serviceRecord Records of services found.
	 * @param macAddress MAC address of Bluetooth device.
	 */
	public void addServiceRecordToDevice(ServiceRecord serviceRecord,
			String macAddress) {
		if (!deviceMap.containsKey(macAddress))
			return;

		AdvertisingDevice advertisingDevice = deviceMap.get(macAddress);
		advertisingDevice.setObexServiceRecord(serviceRecord);
	}

	/**
	 * Remove a device from management.
	 * 
	 * @param macAddress MAC address of device.
	 * @return The removed device, <code>null</code> if not found.
	 */
	public AdvertisingDevice removeDevice(String macAddress) {
		return deviceMap.remove(macAddress);
	}

	/**
	 * Check if device is already managed.
	 * 
	 * @param macAddress MAC address of device.
	 * @return <code>true</code> if device is already managed.
	 */
	public boolean isNewDevice(String macAddress) {
		return (!deviceMap.containsKey(macAddress));
	}

	/**
	 * Remove from management all devices which has not been shown for
	 * some specified time.
	 * 
	 * @param now Current timestamp.
	 * @param seconds Purge devices not found for <code>seconds</code>.
	 * @return Removed devices list.
	 */
	public List<AdvertisingDevice> purgeDevicesNotDiscoveredFor(Date now,
			long seconds) {
		List<AdvertisingDevice> notDiscovered = new LinkedList<AdvertisingDevice>();
		long nowMillis = now.getTime();
		long secondsMillis = seconds * 1000;

		for (AdvertisingDevice device : deviceMap.values()) {
			if ((nowMillis - device.getLastDiscoveredTimestamp().getTime()) > secondsMillis) {
				notDiscovered.add(device);
				deviceMap.remove(device);
			}
		}

		return notDiscovered;
	}

	/**
	 * Check current service search status of device.
	 * 
	 * @param currentMacAddress MAC address of Bluetooth device.
	 * @return Service search status of device.
	 */
	public ServiceStatus getServiceSearchStatus(String currentMacAddress) {
		AdvertisingDevice device = deviceMap.get(currentMacAddress);

		if (device == null)
			return ServiceStatus.INVALID;
		return device.getServiceSearchStatus();
	}

	/**
	 * Check if device has been found on recent discovery processes.
	 * 
	 * @param macAddress Device MAC address.
	 * @return <code>true</code> if device has been found recently.
	 */
	public boolean isDeviceEnabled(String macAddress) {
		AdvertisingDevice device = deviceMap.get(macAddress);

		if (device == null)
			return false;
		return device.isEnabled();
	}

	/**
	 * Get managed device by MAC address.
	 * 
	 * @param macAddress MAC address of device.
	 * @return Managed device, <code>null</code> if not found.
	 */
	public AdvertisingDevice getDevice(String macAddress) {
		return deviceMap.get(macAddress);
	}

	/**
	 * Update service search status for an existing device.
	 * 
	 * @param macAddress Device MAC address.
	 * @param status Service search status.
	 * @param serviceRecord Current device service record.
	 */
	public void updateServiceSearchStatus(String macAddress,
			ServiceStatus status, ServiceRecord[] serviceRecord) {
		AdvertisingDevice device = deviceMap.get(macAddress);
		device.setServiceSearchStatus(status);

		if (serviceRecord != null)
			device.setObexServiceRecord(serviceRecord[0]);

		if ((status != ServiceStatus.COMPLETED)
				|| (status != ServiceStatus.IN_PROGRESS))
			device.incrementServiceSearchFailures();
	}

	/**
	 * Get a list of devices currently being detected by the discovery
	 * process.
	 * 
	 * @return List of devices currently detected by discovery processes.
	 */
	public List<AdvertisingDevice> getEnabledDevices() {
		LinkedList<AdvertisingDevice> enabledDeviceList = new LinkedList<AdvertisingDevice>();

		for (AdvertisingDevice device : deviceMap.values()) {
			if (device.isEnabled())
				enabledDeviceList.add(device);
		}

		return enabledDeviceList;
	}

	/**
	 * Get a list of devices recently detected by discovery processes,
	 * ordered by advertising priority.
	 * 
	 * @param priorityPolicy Priority policy for ordering the result.
	 * @return List of devices ordered by priority policy.
	 */
	public List<AdvertisingDevice> getEnabledDevicesByPriority(
			PriorityPolicy priorityPolicy) {
		List<AdvertisingDevice> enabledDeviceList = getEnabledDevices();
		Collections.sort(enabledDeviceList, priorityPolicy);
		
		return enabledDeviceList;
	}

	/**
	 * Set service search status for an existing device, providing 
	 * directly the connection URL. Used for NFC-found devices.
	 * 
	 * @param macAddress Device MAC address.
	 * @param status Service search status.
	 * @param btConnectionURL JSR 82 connection URL.
	 */
	public void setServiceSearchStatus(String macAddress, ServiceStatus status,
			String btConnectionURL) {
		AdvertisingDevice device = deviceMap.get(macAddress);
		device.setServiceSearchStatus(status);
		device.setConnectionURL(btConnectionURL);
	}

	/**
	 * Update when device has been discovered again by a discovery process.
	 * 
	 * @param macAddress Device MAC address.
	 * @param timestamp When device has been seem by discovery process.
	 */
	public void updateDeviceDiscoveryStatus(String macAddress, Date timestamp) {
		AdvertisingDevice device = deviceMap.get(macAddress);

		device.setLastDiscoveredTimestamp(timestamp);
		device.incrementTimesDiscovered();
	}

}
