package br.ime.usp.mac499.test.bluetooth.common;

import static org.junit.Assert.*;
import org.junit.Before;
import org.junit.Test;

import br.ime.usp.mac499.bluetooth.common.BluetoothAdapterMonitor;

public class BluetoothAdapterMonitorTest {

	private static final int MAX_CONCURRENT_TRANSFERS = 4;

	private static final int MAX_ADAPTERS = 4;

	private BluetoothAdapterMonitor monitorSingleAdapter;

	private BluetoothAdapterMonitor monitorMultipleAdapter;

	@Before
	public void setUpBefore() throws Exception {
		this.monitorSingleAdapter = new BluetoothAdapterMonitor(1, 0,
				MAX_ADAPTERS);
		this.monitorMultipleAdapter = new BluetoothAdapterMonitor(MAX_ADAPTERS,
				0, MAX_CONCURRENT_TRANSFERS);
	}

	@Test
	public void testAcquireAndReleaseAdapterForDiscovery() {
		Thread singleThread;
		Thread multipleThread;
		MonitorConsumer singleConsumer;
		MonitorConsumer multipleConsumer;

		for (int i = 0; i < 32; i++) {
			singleConsumer = new MonitorConsumer(monitorSingleAdapter);
			singleThread = new Thread(singleConsumer);
			singleThread.start();

			singleConsumer.doAcquire();
			assertTrue(singleConsumer.isAcquired());
			singleConsumer.doRelease();
			assertTrue(singleConsumer.isReleased());
			assertTrue(singleConsumer.isFinished());

			multipleConsumer = new MonitorConsumer(monitorMultipleAdapter);
			multipleThread = new Thread(multipleConsumer);
			multipleThread.start();

			multipleConsumer.doAcquire();
			assertTrue(multipleConsumer.isAcquired());
			multipleConsumer.doRelease();
			assertTrue(multipleConsumer.isReleased());
			assertTrue(multipleConsumer.isFinished());
		}
	}

	@Test
	public void testAcquireAndReleaseAdapterForTransfer() {
		for (int i = 0; i < 8 * MAX_ADAPTERS; i++) {
			monitorSingleAdapter.acquireAdapter(0);
			monitorSingleAdapter.releaseAdapter(0);

			for (int j = 0; j < MAX_ADAPTERS; j++) {
				for (int x = 0; x < MAX_CONCURRENT_TRANSFERS; x++)
					monitorMultipleAdapter.acquireAdapter(j);
				for (int y = 0; y < MAX_CONCURRENT_TRANSFERS; y++)
					monitorMultipleAdapter.releaseAdapter(j);
			}
		}
	}

	@Test
	public void testBlockOnAcquireForTransferSingleAdapter()
			throws InterruptedException {
		for (int i = 0; i < MAX_CONCURRENT_TRANSFERS; i++)
			monitorSingleAdapter.acquireAdapter(0);

		MonitorConsumer consumer = new BlockMonitorConsumer(
				monitorSingleAdapter, 0);
		Thread blockThread = new Thread(consumer);
		blockThread.start();

		for (int i = 0; i < 100; i++)
			assertFalse(consumer.isAcquired());
		monitorSingleAdapter.releaseAdapter(0);
		monitorSingleAdapter.releaseAdapter(0);
		Thread.sleep(500);
		assertTrue(consumer.isAcquired());
	}

	@Test
	public void testAcquireForDiscoveryWhileTransferring()
			throws InterruptedException {
		monitorSingleAdapter.acquireAdapter(0);

		MonitorConsumer consumer = new DiscoveryMonitorConsumer(
				monitorSingleAdapter);
		Thread blockThread = new Thread(consumer);
		blockThread.start();

		for (int i = 0; i < 100; i++)
			assertFalse(consumer.isAcquired());
		monitorSingleAdapter.releaseAdapter(0);
		Thread.sleep(500);
		assertTrue(consumer.isAcquired());
	}

	private class MonitorConsumer implements Runnable {

		private BluetoothAdapterMonitor monitor;

		private boolean acquired;

		private boolean released;

		private boolean finished;

		public MonitorConsumer(BluetoothAdapterMonitor monitor) {
			this.monitor = monitor;
			this.acquired = false;
			this.released = false;
			this.finished = false;
		}

		@Override
		public void run() {
			while (!acquired && !released) {
				try {
					Thread.sleep(500);
				} catch (InterruptedException e) {
					e.printStackTrace();
				}
			}
		}

		public boolean isFinished() {
			return finished;
		}

		public boolean isAcquired() {
			return acquired;
		}

		public boolean isReleased() {
			return released;
		}

		public void doAcquire() {
			monitor.acquireAdapterForDiscovery();
			this.acquired = true;
		}

		public void doRelease() {
			monitor.releaseAdapterForDiscovery();
			this.released = true;

			if (this.acquired)
				this.finished = true;
		}

	}

	private class BlockMonitorConsumer extends MonitorConsumer {

		private int adapter;

		public BlockMonitorConsumer(BluetoothAdapterMonitor monitor, int adapter) {
			super(monitor);
			this.adapter = adapter;
		}

		@Override
		public void run() {
			super.monitor.acquireAdapter(adapter);
			super.acquired = true;
		}

	}

	private class DiscoveryMonitorConsumer extends MonitorConsumer {

		public DiscoveryMonitorConsumer(BluetoothAdapterMonitor monitor) {
			super(monitor);
		}

		@Override
		public void run() {
			super.monitor.acquireAdapterForDiscovery();
			super.acquired = true;
		}

	}

}
