package br.ime.usp.mac499.test.bluetooth.model;

import java.text.DateFormat;
import java.text.SimpleDateFormat;
import java.util.Arrays;
import java.util.Date;
import java.util.LinkedList;
import java.util.List;

import static org.junit.Assert.*;
import org.junit.BeforeClass;
import org.junit.Test;

import br.ime.usp.mac499.bluetooth.common.sdp.ServiceStatus;
import br.ime.usp.mac499.bluetooth.io.OBEXPushStatus;
import br.ime.usp.mac499.bluetooth.model.AdvertisementAttempt;
import br.ime.usp.mac499.bluetooth.model.AdvertisingDevice;
import br.ime.usp.mac499.bluetooth.model.DefaultPriorityPolicy;
import br.ime.usp.mac499.bluetooth.model.Priority;
import br.ime.usp.mac499.bluetooth.model.PriorityPolicy;

public class DefaultPriorityPolicyTest {

	private static List<AdvertisingDevice> deviceList;

	@BeforeClass
	public static void setUpBeforeClass() throws Exception {
		deviceList = new LinkedList<AdvertisingDevice>();
		DateFormat dateFormat = new SimpleDateFormat("MM-dd-yyyy HH:mm:ss");

		AdvertisementAttempt succesfulAttempt = new AdvertisementAttempt(null,
				null, null, OBEXPushStatus.SUCCESS);
		AdvertisementAttempt errorAttempt = new AdvertisementAttempt(null,
				null, null, OBEXPushStatus.FAILED);

		String currentDate = "01-01-2010 00:00:00";
		Date parsedDate = dateFormat.parse(currentDate);
		AdvertisingDevice currentDevice = new AdvertisingDevice("1", null,
				parsedDate);
		deviceList.add(currentDevice);

		currentDate = "01-01-2010 00:00:00";
		parsedDate = dateFormat.parse(currentDate);
		currentDevice = new AdvertisingDevice("2", null, parsedDate);
		deviceList.add(currentDevice);

		currentDate = "01-01-2010 00:10:00";
		parsedDate = dateFormat.parse(currentDate);
		currentDevice = new AdvertisingDevice("3", null, parsedDate);
		deviceList.add(currentDevice);

		currentDate = "01-01-2010 00:00:00";
		parsedDate = dateFormat.parse(currentDate);
		currentDevice = new AdvertisingDevice("4", null, parsedDate);
		currentDevice.incrementServiceSearchFailures();
		deviceList.add(currentDevice);

		currentDate = "01-01-2010 00:00:00";
		parsedDate = dateFormat.parse(currentDate);
		currentDevice = new AdvertisingDevice("5", null, parsedDate);
		currentDevice.incrementTimesDiscovered();
		deviceList.add(currentDevice);

		currentDate = "01-01-2010 00:00:00";
		parsedDate = dateFormat.parse(currentDate);
		currentDevice = new AdvertisingDevice("6", null, parsedDate);
		currentDevice.addAdvertisementAttempt(succesfulAttempt);
		deviceList.add(currentDevice);

		currentDate = "01-01-2010 00:00:00";
		parsedDate = dateFormat.parse(currentDate);
		currentDevice = new AdvertisingDevice("7", null, parsedDate);
		currentDevice.addAdvertisementAttempt(errorAttempt);
		deviceList.add(currentDevice);

		currentDate = "01-01-2010 00:30:00";
		parsedDate = dateFormat.parse(currentDate);
		currentDevice = new AdvertisingDevice("8", null, parsedDate);
		for (int i = 0; i < 10; i++) {
			currentDevice.incrementTimesDiscovered();
			currentDevice.addAdvertisementAttempt(succesfulAttempt);
		}
		deviceList.add(currentDevice);

		currentDate = "01-01-2010 00:30:00";
		parsedDate = dateFormat.parse(currentDate);
		currentDevice = new AdvertisingDevice("9", null, parsedDate);
		for (int i = 0; i < 10; i++) {
			currentDevice.incrementTimesDiscovered();
			currentDevice.incrementServiceSearchFailures();
			currentDevice.addAdvertisementAttempt(errorAttempt);
		}
		deviceList.add(currentDevice);

		currentDate = "01-01-2009 00:00:00";
		parsedDate = dateFormat.parse(currentDate);
		currentDevice = new AdvertisingDevice("10", null, parsedDate);
		currentDevice.setNfcInitiated(true);
		currentDevice.setLastDiscoveredTimestamp(parsedDate);
		currentDevice.setServiceSearchStatus(ServiceStatus.COMPLETED);
		deviceList.add(currentDevice);
	}

	@Test
	public void testUselessPolicy() {
		PriorityPolicy policy = new DefaultPriorityPolicy();

		AdvertisingDevice[] deviceArray = getArray(deviceList, policy);
		String[] expectedOrder = { "1", "2", "3", "4", "5", "6", "7", "8", "9",
				"10" };
		performTest(deviceArray, expectedOrder);
	}

	@Test
	public void testTimesDiscoveredPriority() {
		PriorityPolicy policy = new DefaultPriorityPolicy(
				Priority.TIMES_DISCOVERED);

		AdvertisingDevice[] deviceArray = getArray(deviceList, policy);
		String[] expectedOrder = { "8", "9", "5", "1", "2", "3", "4", "6", "7" };
		performTest(deviceArray, expectedOrder);
	}

	@Test
	public void testNfcInitiatedPolicy() {
		PriorityPolicy policy = new DefaultPriorityPolicy(
				Priority.NFC_INITIATED);

		AdvertisingDevice[] deviceArray = getArray(deviceList, policy);
		String[] expectedOrder = { "10", "1", "2", "3", "4", "5", "6", "7",
				"8", "9" };
		performTest(deviceArray, expectedOrder);
	}

	@Test
	public void testServiceSearchErrorPolicy() {
		PriorityPolicy policy = new DefaultPriorityPolicy(
				Priority.SERVICE_SEARCH_ERROR);

		AdvertisingDevice[] deviceArray = getArray(deviceList, policy);
		String[] expectedOrder = { "1", "2", "3", "5", "6", "7", "8", "10",
				"4", "9" };
		performTest(deviceArray, expectedOrder);
	}

	@Test
	public void testAdvertisingSuccessPolicy() {
		PriorityPolicy policy = new DefaultPriorityPolicy(
				Priority.ADVERTISING_SUCCESS);

		AdvertisingDevice[] deviceArray = getArray(deviceList, policy);
		String[] expectedOrder = { "1", "2", "3", "4", "5", "7", "9", "10",
				"6", "8" };
		performTest(deviceArray, expectedOrder);
	}

	@Test
	public void testAdvertisingErrorPolicy() {
		PriorityPolicy policy = new DefaultPriorityPolicy(
				Priority.ADVERTISING_ERROR);

		AdvertisingDevice[] deviceArray = getArray(deviceList, policy);
		String[] expectedOrder = { "1", "2", "3", "4", "5", "6", "8", "10",
				"7", "9" };
		performTest(deviceArray, expectedOrder);
	}

	@Test
	public void testWhenDiscoveredPolicy() {
		PriorityPolicy policy = new DefaultPriorityPolicy(
				Priority.WHEN_DISCOVERED);

		AdvertisingDevice[] deviceArray = getArray(deviceList, policy);
		String[] expectedOrder = { "8", "9", "3", "1", "2", "4", "5", "6", "7" };
		performTest(deviceArray, expectedOrder);
	}

	@Test
	public void testCommonUsePolicy() {
		PriorityPolicy policy = new DefaultPriorityPolicy(
				Priority.NFC_INITIATED, Priority.WHEN_DISCOVERED,
				Priority.TIMES_DISCOVERED, Priority.ADVERTISING_SUCCESS,
				Priority.SERVICE_SEARCH_ERROR);

		AdvertisingDevice[] deviceArray = getArray(deviceList, policy);
		String[] expectedOrder = { "10", "9", "8", "3", "5", "1", "2", "7",
				"4", "6" };
		performTest(deviceArray, expectedOrder);
	}

	private void performTest(AdvertisingDevice[] deviceArray,
			String[] expectedOrder) {
		for (int i = 0; i < expectedOrder.length; i++)
			assertEquals(expectedOrder[i], deviceArray[i].getMacAddress());
	}

	private AdvertisingDevice[] getArray(
			List<AdvertisingDevice> advertisingDevices,
			PriorityPolicy priorityPolicy) {
		AdvertisingDevice[] deviceArray = new AdvertisingDevice[deviceList
				.size()];
		deviceList.toArray(deviceArray);
		Arrays.sort(deviceArray, priorityPolicy);

		return deviceArray;
	}

}
